/******************************************************************************
 *    CameraService.cc
 *
 *    This file is part of PyNumbra
 *    Copyright (C) 2004 Tom N Harris <telliamed@whoopdedo.cjb.net>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#if (_DARKGAME == 2)

#include <Python.h>
#include "pydefs.h"
#include <lg/iiddef.h>
#include <lg/scrservices.h>
#include <lg/types.h>

#include "ScriptServices.h"
#include "PyNumbraExceptions.h"
#include "pyutils.h"

namespace PyNumbra
{

	const char *
	ScriptService<ICameraSrv>::Name = "CameraService";

	const char * 
	ScriptService<ICameraSrv>::DocStrings[] = {
		"CameraService()\n\t"
			"Provides an interface for attaching the game camera to objects\n\t"
			"other than the player. Only available in Thief 2.",
		"StaticAttach(obj) -> Integer\n\t"
			"Fixes the camera to the object. The orientation is locked.\n\t"
			"If the object has a physics model, a lens overlay will be drawn\n\t"
			"on the screen.",
		"DynamicAttach(obj) -> Integer\n\t"
			"Fixes the camera to the object. The object will be rotated\n\t"
			"in response to user input. If the object has a physics model,\n\t"
			"a lens overlay will be drawn on the screen.",
		"CameraReturn(obj) -> Integer\n\t"
			"If the camera is currently attached to the object, it will be\n\t"
			"detached and returned to the Player object.",
		"ForceCameraReturn() -> Integer\n\t"
			"Returns the camera to the Player object.",
	};

	const PyMethodDef 
	ScriptService<ICameraSrv>::Methods[] = {
		{	"StaticAttach",
			reinterpret_cast<PyCFunction>(StaticAttach),
			METH_VARARGS,
			const_cast<char*>(DocStrings[1])
		},
		{	"DynamicAttach",
			reinterpret_cast<PyCFunction>(DynamicAttach),
			METH_VARARGS,
			const_cast<char*>(DocStrings[2])
		},
		{	"CameraReturn",
			reinterpret_cast<PyCFunction>(CameraReturn),
			METH_VARARGS,
			const_cast<char*>(DocStrings[3])
		},
		{	"ForceCameraReturn",
			reinterpret_cast<PyCFunction>(ForceCameraReturn),
			METH_NOARGS,
			const_cast<char*>(DocStrings[4])
		},
		{ NULL, NULL, 0, NULL }
	};

	PyTypeObject 
	ScriptService<ICameraSrv>::TypeInfo = {
		PyObject_HEAD_INIT(NULL)
		0,
		const_cast<char*>(ScriptService<ICameraSrv>::Name),
		sizeof(ScriptServiceObject<ICameraSrv>), 0,
		reinterpret_cast<destructor>(ScriptService<IUnknown>::Dealloc),
		0,
		0, 0,		/* tp_getattr, tp_setattr */
		reinterpret_cast<cmpfunc>(ScriptService<IUnknown>::Compare),
		reinterpret_cast<reprfunc>(ScriptService<IUnknown>::Repr),
		0, 0, 0,	/* tp_as_number, tp_as_sequence, tp_as_mapping */
		0, 0,		/* tp_hash, tp_call */
		0,			/* tp_str */
		0, 0, 0,	/* tp_getattro, tp_setattro, tp_as_buffer */
		Py_TPFLAGS_DEFAULT,		/* tp_flags */
		const_cast<char*>(ScriptService<ICameraSrv>::DocStrings[0]),
		0, 0, 0,	/* tp_traverse, tp_clear, tp_richcompare */
		0, 0, 0,	/* tp_weaklistoffset, tp_iter, tp_iternext */
		const_cast<PyMethodDef*>(ScriptService<ICameraSrv>::Methods),
		0,0,0,0,0,0,0,
		0,0,0,
		0,0,0,0,0,0,0,0
	};


	ScriptServiceObject<ICameraSrv> * 
	ScriptService<ICameraSrv>::NewObject(PyObject *)
	{
		ICameraSrv *iface = static_cast<ICameraSrv*>(g_pScriptManager->GetService(IID_ICameraScriptService));
		if (!iface)
		{
			Exception::Raise(Exception::NoInterfaceError, Name);
			return NULL;
		}
		ScriptServiceObject<ICameraSrv> *obj = PyObject_New(ScriptServiceObject<ICameraSrv>, &TypeInfo);
		if (!obj)
		{
			iface->Release();
			return NULL;
		}
		obj->m_pIFace = iface;
		return obj;
	}

	PyObject * 
	ScriptService<ICameraSrv>::StaticAttach(ScriptServiceObject<ICameraSrv> *self, PyObject *args)
	{
		object obj;
		if (!PyArg_ParseTuple(args, "O&:StaticAttach", cvtObject, &obj))
		{
			return NULL;
		}
		long ret = self->m_pIFace->StaticAttach(obj);
		return PyInt_FromLong(ret);
	}

	PyObject * 
	ScriptService<ICameraSrv>::DynamicAttach(ScriptServiceObject<ICameraSrv> *self, PyObject *args)
	{
		object obj;
		if (!PyArg_ParseTuple(args, "O&:DynamicAttach", cvtObject, &obj))
		{
			return NULL;
		}
		long ret = self->m_pIFace->DynamicAttach(obj);
		return PyInt_FromLong(ret);
	}

	PyObject * 
	ScriptService<ICameraSrv>::CameraReturn(ScriptServiceObject<ICameraSrv> *self, PyObject *args)
	{
		object obj;
		if (!PyArg_ParseTuple(args, "O&:CameraReturn", cvtObject, &obj))
		{
			return NULL;
		}
		long ret = self->m_pIFace->CameraReturn(obj);
		return PyInt_FromLong(ret);
	}

	PyObject * 
	ScriptService<ICameraSrv>::ForceCameraReturn(ScriptServiceObject<ICameraSrv> *self)
	{
		long ret = self->m_pIFace->ForceCameraReturn();
		return PyInt_FromLong(ret);
	}


} // namespace PyNumbra

#endif
