/******************************************************************************
 *    DamageService.cc
 *
 *    This file is part of PyNumbra
 *    Copyright (C) 2004 Tom N Harris <telliamed@whoopdedo.cjb.net>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <Python.h>
#include "pydefs.h"
#include <lg/iiddef.h>
#include <lg/scrservices.h>
#include <lg/types.h>

#include "ScriptServices.h"
#include "PyNumbraExceptions.h"
#include "pyutils.h"

namespace PyNumbra
{

	const char *
	ScriptService<IDamageSrv>::Name = "DamageService";

	const char * 
	ScriptService<IDamageSrv>::DocStrings[] = {
		"DamageService()\n\t"
			"Bestow unimaginable carnage upon all those who oppose you.",
		"Damage(victim,culprit,intensity,type) -> Integer\n\t"
			"Causes intensity points of damage to the victim by the culprit.\n\t"
			"type is an integer.",
		"Slay(victim,culprit) -> Integer\n\t"
			"Makes the victim dead. Hit points are unaffected.",
		"Resurrect(unvictim,culprit) -> Integer\n\t"
			"Makes the unvictim not-dead. Hit points are unaffected.",
	};

	const PyMethodDef 
	ScriptService<IDamageSrv>::Methods[] = {
		{	"Damage",
			reinterpret_cast<PyCFunction>(Damage),
			METH_VARARGS,
			const_cast<char*>(DocStrings[1])
		},
		{	"Slay",
			reinterpret_cast<PyCFunction>(Slay),
			METH_VARARGS,
			const_cast<char*>(DocStrings[2])
		},
		{	"Resurrect",
			reinterpret_cast<PyCFunction>(Resurrect),
			METH_VARARGS,
			const_cast<char*>(DocStrings[3])
		},
		{ NULL, NULL, 0, NULL }
	};

	PyTypeObject 
	ScriptService<IDamageSrv>::TypeInfo = {
		PyObject_HEAD_INIT(NULL)
		0,
		const_cast<char*>(ScriptService<IDamageSrv>::Name),
		sizeof(ScriptServiceObject<IDamageSrv>), 0,
		reinterpret_cast<destructor>(ScriptService<IUnknown>::Dealloc),
		0,
		0, 0,		/* tp_getattr, tp_setattr */
		reinterpret_cast<cmpfunc>(ScriptService<IUnknown>::Compare),
		reinterpret_cast<reprfunc>(ScriptService<IUnknown>::Repr),
		0, 0, 0,	/* tp_as_number, tp_as_sequence, tp_as_mapping */
		0, 0,		/* tp_hash, tp_call */
		0,			/* tp_str */
		0, 0, 0,	/* tp_getattro, tp_setattro, tp_as_buffer */
		Py_TPFLAGS_DEFAULT,		/* tp_flags */
		const_cast<char*>(ScriptService<IDamageSrv>::DocStrings[0]),
		0, 0, 0,	/* tp_traverse, tp_clear, tp_richcompare */
		0, 0, 0,	/* tp_weaklistoffset, tp_iter, tp_iternext */
		const_cast<PyMethodDef*>(ScriptService<IDamageSrv>::Methods),
		0,0,0,0,0,0,0,
		0,0,0,
		0,0,0,0,0,0,0,0
	};


	ScriptServiceObject<IDamageSrv> * 
	ScriptService<IDamageSrv>::NewObject(PyObject *)
	{
		IDamageSrv *iface = static_cast<IDamageSrv*>(g_pScriptManager->GetService(IID_IDamageScriptService));
		if (!iface)
		{
			Exception::Raise(Exception::NoInterfaceError, Name);
			return NULL;
		}
		ScriptServiceObject<IDamageSrv> *obj = PyObject_New(ScriptServiceObject<IDamageSrv>, &TypeInfo);
		if (!obj)
		{
			iface->Release();
			return NULL;
		}
		obj->m_pIFace = iface;
		return obj;
	}

	PyObject * 
	ScriptService<IDamageSrv>::Damage(ScriptServiceObject<IDamageSrv> *self, PyObject *args)
	{
		object victim, culprit;
		int intensity, type;
		if (!PyArg_ParseTuple(args, "O&O&ii:Damage", cvtObject, &victim, cvtObject, &culprit, &intensity, &type))
		{
			return NULL;
		}
		long ret = self->m_pIFace->Damage(victim,culprit,intensity,type);
		return PyInt_FromLong(ret);
	}

	PyObject * 
	ScriptService<IDamageSrv>::Slay(ScriptServiceObject<IDamageSrv> *self, PyObject *args)
	{
		object victim, culprit;
		if (!PyArg_ParseTuple(args, "O&O&:Slay", cvtObject, &victim, cvtObject, &culprit))
		{
			return NULL;
		}
		long ret = self->m_pIFace->Slay(victim,culprit);
		return PyInt_FromLong(ret);
	}

	PyObject * 
	ScriptService<IDamageSrv>::Resurrect(ScriptServiceObject<IDamageSrv> *self, PyObject *args)
	{
		object unvictim, culprit;
		if (!PyArg_ParseTuple(args, "O&O&:Resurrect", cvtObject, &unvictim, cvtObject, &culprit))
		{
			return NULL;
		}
		long ret = self->m_pIFace->Resurrect(unvictim,culprit);
		return PyInt_FromLong(ret);
	}


} // namespace PyNumbra


