/******************************************************************************
 *    DataService.cc
 *
 *    This file is part of PyNumbra
 *    Copyright (C) 2004 Tom N Harris <telliamed@whoopdedo.cjb.net>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <Python.h>
#include "pydefs.h"
#include <lg/iiddef.h>
#include <lg/scrservices.h>
#include <lg/types.h>

#include "ScriptServices.h"
#include "PyNumbraExceptions.h"
#include "pyutils.h"

namespace PyNumbra
{

	const char *
	ScriptService<IDataSrv>::Name = "DataService";

	const char * 
	ScriptService<IDataSrv>::DocStrings[] = {
		"DataService()\n\t"
			"Generic functions related to strings and numbers.",
		"GetString(resfile,strname,unused=\"\",res=\"strings\") -> String\n\t"
			"If you want to retrieve a book string, you can't use res='books'.\n\t"
			"You must, instead, use res='strings' and prefix resfile with\n\t"
			"'../books/'.\n\t"
			"I've never seen the third argument used.",
		"GetObjString(obj,type) -> String\n\t"
			"Retrieves a string associated with the object. type describes the\n\t"
			"string that will be retrieved.\n\t"
			"Thief and Thief2 recognize 'objnames' and 'objdescs'.\n\t"
			"System Shock 2 recognizes 'ObjName', 'ObjShort', 'ObjLookS',\n\t"
			"'Modify1', 'Modify2', 'HackText', 'UseMsg', 'LockMsg', 'Sett1',\n\t"
			"'Sett2', 'SHead1', 'SHead2', 'RsrchTxt', 'HUDUse', 'WorldCursor',\n\t"
			"'InvCursor', 'UseCursor'.",
		"DirectRand() -> Integer\n\t"
			"[0,32767]",
		"RandInt(low,high) -> Integer\n\t"
			"[low,high]",
		"RandFlt0to1() -> Float\n\t"
			"[0.0,1.0]",
		"RandFltNeg1to1() -> Float\n\t"
			"(-1.0,1.0]",
	};

	const PyMethodDef 
	ScriptService<IDataSrv>::Methods[] = {
		{	"GetString",
			reinterpret_cast<PyCFunction>(GetString),
			METH_VARARGS,
			const_cast<char*>(DocStrings[1])
		},
		{	"GetObjString",
			reinterpret_cast<PyCFunction>(GetObjString),
			METH_VARARGS,
			const_cast<char*>(DocStrings[2])
		},
		{	"DirectRand",
			reinterpret_cast<PyCFunction>(DirectRand),
			METH_NOARGS,
			const_cast<char*>(DocStrings[3])
		},
		{	"RandInt",
			reinterpret_cast<PyCFunction>(RandInt),
			METH_VARARGS,
			const_cast<char*>(DocStrings[4])
		},
		{	"RandFlt0to1",
			reinterpret_cast<PyCFunction>(RandFlt0to1),
			METH_NOARGS,
			const_cast<char*>(DocStrings[5])
		},
		{	"RandFltNeg1to1",
			reinterpret_cast<PyCFunction>(RandFltNeg1to1),
			METH_NOARGS,
			const_cast<char*>(DocStrings[6])
		},
		{ NULL, NULL, 0, NULL }
	};

	PyTypeObject 
	ScriptService<IDataSrv>::TypeInfo = {
		PyObject_HEAD_INIT(NULL)
		0,
		const_cast<char*>(ScriptService<IDataSrv>::Name),
		sizeof(ScriptServiceObject<IDataSrv>), 0,
		reinterpret_cast<destructor>(ScriptService<IUnknown>::Dealloc),
		0,
		0, 0,		/* tp_getattr, tp_setattr */
		reinterpret_cast<cmpfunc>(ScriptService<IUnknown>::Compare),
		reinterpret_cast<reprfunc>(ScriptService<IUnknown>::Repr),
		0, 0, 0,	/* tp_as_number, tp_as_sequence, tp_as_mapping */
		0, 0,		/* tp_hash, tp_call */
		0,			/* tp_str */
		0, 0, 0,	/* tp_getattro, tp_setattro, tp_as_buffer */
		Py_TPFLAGS_DEFAULT,		/* tp_flags */
		const_cast<char*>(ScriptService<IDataSrv>::DocStrings[0]),
		0, 0, 0,	/* tp_traverse, tp_clear, tp_richcompare */
		0, 0, 0,	/* tp_weaklistoffset, tp_iter, tp_iternext */
		const_cast<PyMethodDef*>(ScriptService<IDataSrv>::Methods),
		0,0,0,0,0,0,0,
		0,0,0,
		0,0,0,0,0,0,0,0
	};


	ScriptServiceObject<IDataSrv> * 
	ScriptService<IDataSrv>::NewObject(PyObject *)
	{
		IDataSrv *iface = static_cast<IDataSrv*>(g_pScriptManager->GetService(IID_IDataScriptService));
		if (!iface)
		{
			Exception::Raise(Exception::NoInterfaceError, Name);
			return NULL;
		}
		ScriptServiceObject<IDataSrv> *obj = PyObject_New(ScriptServiceObject<IDataSrv>, &TypeInfo);
		if (!obj)
		{
			iface->Release();
			return NULL;
		}
		obj->m_pIFace = iface;
		return obj;
	}

	PyObject * 
	ScriptService<IDataSrv>::GetString(ScriptServiceObject<IDataSrv> *self, PyObject *args)
	{
		const char *resfile, *strname;
		const char *lang = "", *res = "strings";
		if (!PyArg_ParseTuple(args, "ss|ss:GetString", &resfile, &strname, &lang, &res))
		{
			return NULL;
		}
		cScrStr str;
		self->m_pIFace->GetString(str,resfile,strname,lang,res);
		PyObject *ret = PyString_FromString(static_cast<const char*>(str));
		str.Free();
		return ret;
	}

	PyObject * 
	ScriptService<IDataSrv>::GetObjString(ScriptServiceObject<IDataSrv> *self, PyObject *args)
	{
		object obj;
		const char *type;
		if (!PyArg_ParseTuple(args, "O&s:GetObjString", cvtObject, &obj, &type))
		{
			return NULL;
		}
		cScrStr str;
		self->m_pIFace->GetObjString(str,obj,type);
		PyObject *ret = PyString_FromString(static_cast<const char*>(str));
		str.Free();
		return ret;
	}

	PyObject * 
	ScriptService<IDataSrv>::DirectRand(ScriptServiceObject<IDataSrv> *self)
	{
		return PyInt_FromLong(self->m_pIFace->DirectRand());
	}

	PyObject * 
	ScriptService<IDataSrv>::RandInt(ScriptServiceObject<IDataSrv> *self, PyObject *args)
	{
		int low,high;
		if (!PyArg_ParseTuple(args, "ii:RandInt", &low, &high))
		{
			return NULL;
		}
		int ret = self->m_pIFace->RandInt(low,high);
		return PyInt_FromLong(ret);
	}

	PyObject * 
	ScriptService<IDataSrv>::RandFlt0to1(ScriptServiceObject<IDataSrv> *self)
	{
		return PyFloat_FromDouble(self->m_pIFace->RandFlt0to1());
	}

	PyObject * 
	ScriptService<IDataSrv>::RandFltNeg1to1(ScriptServiceObject<IDataSrv> *self)
	{
		return PyFloat_FromDouble(self->m_pIFace->RandFltNeg1to1());
	}


} // namespace PyNumbra


