/******************************************************************************
 *    PhysicsService.cc
 *
 *    This file is part of PyNumbra
 *    Copyright (C) 2004 Tom N Harris <telliamed@whoopdedo.cjb.net>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <Python.h>
#include "pydefs.h"
#include <lg/iiddef.h>
#include <lg/scrservices.h>
#include <lg/types.h>

#include "ScriptServices.h"
#include "PyNumbraExceptions.h"
#include "pyutils.h"
#include "PyNumbraTypes.h"

namespace PyNumbra
{

	const char *
	ScriptService<IPhysSrv>::Name = "PhysicsService";

	const char * 
	ScriptService<IPhysSrv>::DocStrings[] = {
		"PhysicsService()\n\t"
			"Things that go bump in the worldrep.",
		"SubscribeMsg(obj,type) -> Integer\n\t"
			"Registers this object to receive one or more of the physics\n\t"
			"messages that occur to the object.",
		"UnsubscribeMsg(obj,type) -> Integer\n\t"
			"Stops listening for the types of physics messages.",
		"LaunchProjectile(src,archetype,speed,num,vec) -> PyNumbra.Object",
		"SetVelocity(obj,vec) -> Integer\n\t"
			"Changes the directional speed of the object.",
		"GetVelocity(obj) -> Vector\n\t"
			"Returns a 3-tuple of the object's speed.",
		"ControlVelocity(obj,vec) -> Integer\n\t"
			"Only available in Thief 2.",
		"StopControlVelocity(obj) -> Integer\n\t"
			"Only available in Thief 2.",
		"SetGravity(obj,gravity) -> Integer\n\t"
			"Changes the amount of gravity the object is subject to.",
		"GetGravity(obj) -> Float\n\t"
			"Returns the object's gravity as a percentage.",
#if (_DARKGAME != 1)
		"HasPhysics(obj) -> Boolean\n\t"
			"Returns True if the object has a valid physics model.\n\t"
			"Only available in Thief 2 and System Shock 2.",
		"IsSphere(obj) -> Boolean\n\t"
			"Checks if the object has sphere-like physics.\n\t"
			"Only available in Thief 2 and System Shock 2.",
		"IsOBB(obj) -> Boolean\n\t"
			"Checks if the object is an oriented bounding-box.\n\t"
			"Only available in Thief 2 and System Shock 2.",
		"ControlCurrentLocation(obj) -> Integer\n\t"
			"Only available in Thief 2 and System Shock 2.",
		"ControlCurrentRotation(obj) -> Integer\n\t"
			"Only available in Thief 2 and System Shock 2.",
		"ControlCurrentPosition(obj) -> Integer\n\t"
			"Only available in Thief 2 and System Shock 2.",
		"DeregisterModel(obj) -> Integer\n\t"
			"Only available in Thief 2 and System Shock 2.",
		"PlayerMotionSetOffset(i,v)\n\t"
			"Only available in Thief 2 and System Shock 2.",
		"Activate(obj) -> Integer\n\t"
			"Only available in Thief 2 and System Shock 2.",
		"ValidPos(obj) -> Boolean\n\t"
			"Only available in Thief 2 and System Shock 2.",
#endif
	};

	const PyMethodDef 
	ScriptService<IPhysSrv>::Methods[] = {
		{	"SubscribeMsg",
			reinterpret_cast<PyCFunction>(SubscribeMsg),
			METH_VARARGS,
			const_cast<char*>(DocStrings[1])
		},
		{	"UnsubscribeMsg",
			reinterpret_cast<PyCFunction>(UnsubscribeMsg),
			METH_VARARGS,
			const_cast<char*>(DocStrings[2])
		},
		{	"LaunchProjectile",
			reinterpret_cast<PyCFunction>(LaunchProjectile),
			METH_VARARGS,
			const_cast<char*>(DocStrings[3])
		},
		{	"SetVelocity",
			reinterpret_cast<PyCFunction>(SetVelocity),
			METH_VARARGS,
			const_cast<char*>(DocStrings[4])
		},
		{	"GetVelocity",
			reinterpret_cast<PyCFunction>(GetVelocity),
			METH_VARARGS,
			const_cast<char*>(DocStrings[5])
		},
#if (_DARKGAME == 2)
		{	"ControlVelocity",
			reinterpret_cast<PyCFunction>(ControlVelocity),
			METH_VARARGS,
			const_cast<char*>(DocStrings[6])
		},
		{	"StopControlVelocity",
			reinterpret_cast<PyCFunction>(StopControlVelocity),
			METH_VARARGS,
			const_cast<char*>(DocStrings[7])
		},
#endif
		{	"SetGravity",
			reinterpret_cast<PyCFunction>(SetGravity),
			METH_VARARGS,
			const_cast<char*>(DocStrings[8])
		},
		{	"GetGravity",
			reinterpret_cast<PyCFunction>(GetGravity),
			METH_VARARGS,
			const_cast<char*>(DocStrings[9])
		},
#if (_DARKGAME != 1)
		{	"HasPhysics",
			reinterpret_cast<PyCFunction>(HasPhysics),
			METH_VARARGS,
			const_cast<char*>(DocStrings[10])
		},
		{	"IsSphere",
			reinterpret_cast<PyCFunction>(IsSphere),
			METH_VARARGS,
			const_cast<char*>(DocStrings[11])
		},
		{	"IsOBB",
			reinterpret_cast<PyCFunction>(IsOBB),
			METH_VARARGS,
			const_cast<char*>(DocStrings[12])
		},
		{	"ControlCurrentLocation",
			reinterpret_cast<PyCFunction>(ControlCurrentLocation),
			METH_VARARGS,
			const_cast<char*>(DocStrings[13])
		},
		{	"ControlCurrentRotation",
			reinterpret_cast<PyCFunction>(ControlCurrentRotation),
			METH_VARARGS,
			const_cast<char*>(DocStrings[14])
		},
		{	"ControlCurrentPosition",
			reinterpret_cast<PyCFunction>(ControlCurrentPosition),
			METH_VARARGS,
			const_cast<char*>(DocStrings[15])
		},
		{	"DeregisterModel",
			reinterpret_cast<PyCFunction>(DeregisterModel),
			METH_VARARGS,
			const_cast<char*>(DocStrings[16])
		},
		{	"PlayerMotionSetOffset",
			reinterpret_cast<PyCFunction>(PlayerMotionSetOffset),
			METH_VARARGS,
			const_cast<char*>(DocStrings[17])
		},
		{	"Activate",
			reinterpret_cast<PyCFunction>(Activate),
			METH_VARARGS,
			const_cast<char*>(DocStrings[18])
		},
		{	"ValidPos",
			reinterpret_cast<PyCFunction>(ValidPos),
			METH_VARARGS,
			const_cast<char*>(DocStrings[19])
		},
#endif
		{ NULL, NULL, 0, NULL }
	};

	PyTypeObject 
	ScriptService<IPhysSrv>::TypeInfo = {
		PyObject_HEAD_INIT(NULL)
		0,
		const_cast<char*>(ScriptService<IPhysSrv>::Name),
		sizeof(ScriptServiceObject<IPhysSrv>), 0,
		reinterpret_cast<destructor>(ScriptService<IUnknown>::Dealloc),
		0,
		0, 0,		/* tp_getattr, tp_setattr */
		reinterpret_cast<cmpfunc>(ScriptService<IUnknown>::Compare),
		reinterpret_cast<reprfunc>(ScriptService<IUnknown>::Repr),
		0, 0, 0,	/* tp_as_number, tp_as_sequence, tp_as_mapping */
		0, 0,		/* tp_hash, tp_call */
		0,			/* tp_str */
		0, 0, 0,	/* tp_getattro, tp_setattro, tp_as_buffer */
		Py_TPFLAGS_DEFAULT,		/* tp_flags */
		const_cast<char*>(ScriptService<IPhysSrv>::DocStrings[0]),
		0, 0, 0,	/* tp_traverse, tp_clear, tp_richcompare */
		0, 0, 0,	/* tp_weaklistoffset, tp_iter, tp_iternext */
		const_cast<PyMethodDef*>(ScriptService<IPhysSrv>::Methods),
		0,0,0,0,0,0,0,
		0,0,0,
		0,0,0,0,0,0,0,0
	};


	void
	ScriptService<IPhysSrv>::InitModule(PyObject *module)
	{
		PyType_Ready(&TypeInfo);
		PyModule_AddIntConstant(module, "kPhysCollision", kPhysCollision);
		PyModule_AddIntConstant(module, "kPhysContact", kPhysContact);
		PyModule_AddIntConstant(module, "kPhysEnterExit", kPhysEnterExit);
	}

	ScriptServiceObject<IPhysSrv> * 
	ScriptService<IPhysSrv>::NewObject(PyObject *)
	{
		IPhysSrv *iface = static_cast<IPhysSrv*>(g_pScriptManager->GetService(IID_IPhysicsScriptService));
		if (!iface)
		{
			Exception::Raise(Exception::NoInterfaceError, Name);
			return NULL;
		}
		ScriptServiceObject<IPhysSrv> *obj = PyObject_New(ScriptServiceObject<IPhysSrv>, &TypeInfo);
		if (!obj)
		{
			iface->Release();
			return NULL;
		}
		obj->m_pIFace = iface;
		return obj;
	}

	PyObject * 
	ScriptService<IPhysSrv>::SubscribeMsg(ScriptServiceObject<IPhysSrv> *self, PyObject *args)
	{
		int type;
		object obj;
		if (!PyArg_ParseTuple(args, "O&i:SubscribeMsg", cvtObject, &obj, &type))
		{
			return NULL;
		}
		long ret = self->m_pIFace->SubscribeMsg(obj,type);
		return PyInt_FromLong(ret);
	}

	PyObject * 
	ScriptService<IPhysSrv>::UnsubscribeMsg(ScriptServiceObject<IPhysSrv> *self, PyObject *args)
	{
		int type;
		object obj;
		if (!PyArg_ParseTuple(args, "O&i:UnsubscribeMsg", cvtObject, &obj, &type))
		{
			return NULL;
		}
		long ret = self->m_pIFace->UnsubscribeMsg(obj,type);
		return PyInt_FromLong(ret);
	}

	PyObject * 
	ScriptService<IPhysSrv>::LaunchProjectile(ScriptServiceObject<IPhysSrv> *self, PyObject *args)
	{
		object src, archetype;
		int num;
		float speed;
		cScrVec vec;
		if (!PyArg_ParseTuple(args, "O&O&fi(fff):LaunchProjectile", cvtObject, &src, cvtObject, &archetype,
							&speed, &num, &vec.pos.x, &vec.pos.y, &vec.pos.z))
		{
			return NULL;
		}
		object ret;
		self->m_pIFace->LaunchProjectile(ret,src,archetype,speed,num,vec);
		if (ret == 0)
		{
			Py_INCREF(Py_None);
			return Py_None;
		}
		return ObjectObject::New(ret);
	}

	PyObject * 
	ScriptService<IPhysSrv>::SetVelocity(ScriptServiceObject<IPhysSrv> *self, PyObject *args)
	{
		object obj;
		cScrVec vec;
		if (!PyArg_ParseTuple(args, "O&(fff):SetVelocity", cvtObject, &obj, &vec.pos.x, &vec.pos.y, &vec.pos.z))
		{
			return NULL;
		}
		long ret = self->m_pIFace->SetVelocity(obj,vec);
		return PyInt_FromLong(ret);
	}

	PyObject * 
	ScriptService<IPhysSrv>::GetVelocity(ScriptServiceObject<IPhysSrv> *self, PyObject *args)
	{
		object obj;
		if (!PyArg_ParseTuple(args, "O&:GetVelocity", cvtObject, &obj))
		{
			return NULL;
		}
		cScrVec vec;
		self->m_pIFace->GetVelocity(obj, vec);
		return Py_BuildValue("[fff]", vec.pos.x, vec.pos.y, vec.pos.z);
	}

#if (_DARKGAME == 2)
	PyObject * 
	ScriptService<IPhysSrv>::ControlVelocity(ScriptServiceObject<IPhysSrv> *self, PyObject *args)
	{
		object obj;
		cScrVec vec;
		if (!PyArg_ParseTuple(args, "O&(fff):ControlVelocity", cvtObject, &obj, &vec.pos.x, &vec.pos.y, &vec.pos.z))
		{
			return NULL;
		}
		long ret = self->m_pIFace->ControlVelocity(obj,vec);
		return PyInt_FromLong(ret);
	}

	PyObject * 
	ScriptService<IPhysSrv>::StopControlVelocity(ScriptServiceObject<IPhysSrv> *self, PyObject *args)
	{
		object obj;
		if (!PyArg_ParseTuple(args, "O&:StopControlVelocity", cvtObject, &obj))
		{
			return NULL;
		}
		long ret = self->m_pIFace->StopControlVelocity(obj);
		return PyInt_FromLong(ret);
	}
#endif

	PyObject * 
	ScriptService<IPhysSrv>::SetGravity(ScriptServiceObject<IPhysSrv> *self, PyObject *args)
	{
		object obj;
		float gravity;
		if (!PyArg_ParseTuple(args, "O&f:SetGravity", cvtObject, &obj, &gravity))
		{
			return NULL;
		}
		long ret = self->m_pIFace->SetGravity(obj,gravity);
		return PyInt_FromLong(ret);
	}

	PyObject * 
	ScriptService<IPhysSrv>::GetGravity(ScriptServiceObject<IPhysSrv> *self, PyObject *args)
	{
		object obj;
		if (!PyArg_ParseTuple(args, "O&:GetGravity", cvtObject, &obj))
		{
			return NULL;
		}
		float ret = self->m_pIFace->GetGravity(obj);
		return PyFloat_FromDouble(ret);
	}

#if (_DARKGAME != 1)
	PyObject * 
	ScriptService<IPhysSrv>::HasPhysics(ScriptServiceObject<IPhysSrv> *self, PyObject *args)
	{
		object obj;
		if (!PyArg_ParseTuple(args, "O&:HasPhysics", cvtObject, &obj))
		{
			return NULL;
		}
		PyObject *ret = (self->m_pIFace->HasPhysics(obj)) ? Py_True : Py_False;
		Py_INCREF(ret);
		return ret;
	}

	PyObject * 
	ScriptService<IPhysSrv>::IsSphere(ScriptServiceObject<IPhysSrv> *self, PyObject *args)
	{
		object obj;
		if (!PyArg_ParseTuple(args, "O&:IsSphere", cvtObject, &obj))
		{
			return NULL;
		}
		PyObject *ret = (self->m_pIFace->IsSphere(obj)) ? Py_True : Py_False;
		Py_INCREF(ret);
		return ret;
	}

	PyObject * 
	ScriptService<IPhysSrv>::IsOBB(ScriptServiceObject<IPhysSrv> *self, PyObject *args)
	{
		object obj;
		if (!PyArg_ParseTuple(args, "O&:IsOBB", cvtObject, &obj))
		{
			return NULL;
		}
		PyObject *ret = (self->m_pIFace->IsOBB(obj)) ? Py_True : Py_False;
		Py_INCREF(ret);
		return ret;
	}

	PyObject * 
	ScriptService<IPhysSrv>::ControlCurrentLocation(ScriptServiceObject<IPhysSrv> *self, PyObject *args)
	{
		object obj;
		if (!PyArg_ParseTuple(args, "O&:ControlCurrentLocation", cvtObject, &obj))
		{
			return NULL;
		}
		long ret = self->m_pIFace->ControlCurrentLocation(obj);
		return PyInt_FromLong(ret);
	}

	PyObject * 
	ScriptService<IPhysSrv>::ControlCurrentRotation(ScriptServiceObject<IPhysSrv> *self, PyObject *args)
	{
		object obj;
		if (!PyArg_ParseTuple(args, "O&:ControlCurrentRotation", cvtObject, &obj))
		{
			return NULL;
		}
		long ret = self->m_pIFace->ControlCurrentRotation(obj);
		return PyInt_FromLong(ret);
	}

	PyObject * 
	ScriptService<IPhysSrv>::ControlCurrentPosition(ScriptServiceObject<IPhysSrv> *self, PyObject *args)
	{
		object obj;
		if (!PyArg_ParseTuple(args, "O&:ControlCurrentPosition", cvtObject, &obj))
		{
			return NULL;
		}
		long ret = self->m_pIFace->ControlCurrentPosition(obj);
		return PyInt_FromLong(ret);
	}

	PyObject * 
	ScriptService<IPhysSrv>::DeregisterModel(ScriptServiceObject<IPhysSrv> *self, PyObject *args)
	{
		object obj;
		if (!PyArg_ParseTuple(args, "O&:DeregisterModel", cvtObject, &obj))
		{
			return NULL;
		}
		long ret = self->m_pIFace->DeregisterModel(obj);
		return PyInt_FromLong(ret);
	}

	PyObject * 
	ScriptService<IPhysSrv>::PlayerMotionSetOffset(ScriptServiceObject<IPhysSrv> *self, PyObject *args)
	{
		int i;
		cScrVec v;
		if (!PyArg_ParseTuple(args, "i(fff):PlayerMotionSetOffset", &i, &v.pos.x, &v.pos.y, &v.pos.z))
		{
			return NULL;
		}
		self->m_pIFace->PlayerMotionSetOffset(i,v);
		Py_INCREF(Py_None);
		return Py_None;
	}

	PyObject * 
	ScriptService<IPhysSrv>::Activate(ScriptServiceObject<IPhysSrv> *self, PyObject *args)
	{
		object obj;
		if (!PyArg_ParseTuple(args, "O&:Activate", cvtObject, &obj))
		{
			return NULL;
		}
		long ret = self->m_pIFace->Activate(obj);
		return PyInt_FromLong(ret);
	}

	PyObject * 
	ScriptService<IPhysSrv>::ValidPos(ScriptServiceObject<IPhysSrv> *self, PyObject *args)
	{
		object obj;
		if (!PyArg_ParseTuple(args, "O&:ValidPos", cvtObject, &obj))
		{
			return NULL;
		}
		PyObject *ret = (self->m_pIFace->ValidPos(obj)) ? Py_True : Py_False;
		Py_INCREF(ret);
		return ret;
	}
#endif


} // namespace PyNumbra


