/******************************************************************************
 *    PropertyService.cc
 *
 *    This file is part of PyNumbra
 *    Copyright (C) 2004 Tom N Harris <telliamed@whoopdedo.cjb.net>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <Python.h>
#include "pydefs.h"
#include <lg/iiddef.h>
#include <lg/scrservices.h>
#include <lg/types.h>

#include "ScriptServices.h"
#include "PyNumbraExceptions.h"
#include "pyutils.h"

namespace PyNumbra
{

	const char *
	ScriptService<IPropertySrv>::Name = "PropertyService";

	const char * 
	ScriptService<IPropertySrv>::DocStrings[] = {
		"PropertyService()\n\t"
			"Provides an interface for manipulating an object's properties.",
		"Get(obj,name,field=None) -> Integer, Float, String, Vector, or Boolean\n\t"
			"Retrieves a property value.",
		"Set(obj,name,field,data) -> Integer\n\t"
			"Modifies a property value.",
		"SetLocal(obj,name,field,data) -> Integer\n\t"
			"Modifies a property value without propogating it across the\n\t"
			"network.\n\t"
			"Only available in Thief 2 or System Shock 2.",
		"Add(obj,name) -> Integer\n\t"
			"Associate a property with the given object.",
		"Remove(obj,name) -> Integer\n\t"
			"Deletes a property from the object.",
		"CopyFrom(obj,name,src) -> Integer\n\t"
			"Sets the value of a property on obj to match that of src.\n\t"
			"Will add the property if it doesn't already exist on obj.",
		"Possessed(obj,name) -> Boolean\n\t"
			"Checks if the object has a property; includes inheritance.",
	};

	const PyMethodDef 
	ScriptService<IPropertySrv>::Methods[] = {
		{	"Get",
			reinterpret_cast<PyCFunction>(Get),
			METH_VARARGS,
			const_cast<char*>(DocStrings[1])
		},
		{	"Set",
			reinterpret_cast<PyCFunction>(Set),
			METH_VARARGS,
			const_cast<char*>(DocStrings[2])
		},
#if (_DARKGAME == 2) || (_DARKGAME == 3)
		{	"SetLocal",
			reinterpret_cast<PyCFunction>(SetLocal),
			METH_VARARGS,
			const_cast<char*>(DocStrings[3])
		},
#endif
		{	"Add",
			reinterpret_cast<PyCFunction>(Add),
			METH_VARARGS,
			const_cast<char*>(DocStrings[4])
		},
		{	"Remove",
			reinterpret_cast<PyCFunction>(Remove),
			METH_VARARGS,
			const_cast<char*>(DocStrings[5])
		},
		{	"CopyFrom",
			reinterpret_cast<PyCFunction>(CopyFrom),
			METH_VARARGS,
			const_cast<char*>(DocStrings[6])
		},
		{	"Possessed",
			reinterpret_cast<PyCFunction>(Possessed),
			METH_VARARGS,
			const_cast<char*>(DocStrings[7])
		},
		{ NULL, NULL, 0, NULL }
	};

	PyTypeObject 
	ScriptService<IPropertySrv>::TypeInfo = {
		PyObject_HEAD_INIT(NULL)
		0,
		const_cast<char*>(ScriptService<IPropertySrv>::Name),
		sizeof(ScriptServiceObject<IPropertySrv>), 0,
		reinterpret_cast<destructor>(ScriptService<IUnknown>::Dealloc),
		0,
		0, 0,		/* tp_getattr, tp_setattr */
		reinterpret_cast<cmpfunc>(ScriptService<IUnknown>::Compare),
		reinterpret_cast<reprfunc>(ScriptService<IUnknown>::Repr),
		0, 0, 0,	/* tp_as_number, tp_as_sequence, tp_as_mapping */
		0, 0,		/* tp_hash, tp_call */
		0,			/* tp_str */
		0, 0, 0,	/* tp_getattro, tp_setattro, tp_as_buffer */
		Py_TPFLAGS_DEFAULT,		/* tp_flags */
		const_cast<char*>(ScriptService<IPropertySrv>::DocStrings[0]),
		0, 0, 0,	/* tp_traverse, tp_clear, tp_richcompare */
		0, 0, 0,	/* tp_weaklistoffset, tp_iter, tp_iternext */
		const_cast<PyMethodDef*>(ScriptService<IPropertySrv>::Methods),
		0,0,0,0,0,0,0,
		0,0,0,
		0,0,0,0,0,0,0,0
	};


	ScriptServiceObject<IPropertySrv> * 
	ScriptService<IPropertySrv>::NewObject(PyObject *)
	{
		IPropertySrv *iface = static_cast<IPropertySrv*>(g_pScriptManager->GetService(IID_IPropertyScriptService));
		if (!iface)
		{
			Exception::Raise(Exception::NoInterfaceError, Name);
			return NULL;
		}
		ScriptServiceObject<IPropertySrv> *obj = PyObject_New(ScriptServiceObject<IPropertySrv>, &TypeInfo);
		if (!obj)
		{
			iface->Release();
			return NULL;
		}
		obj->m_pIFace = iface;
		return obj;
	}

	PyObject * 
	ScriptService<IPropertySrv>::Get(ScriptServiceObject<IPropertySrv> *self, PyObject *args)
	{
		const char *name, *field = NULL;
		object obj;
		if (!PyArg_ParseTuple(args, "O&s|z:Get", cvtObject, &obj, &name, &field))
		{
			return NULL;
		}
		cMultiParm data;
		self->m_pIFace->Get(data,obj,name,field);
		PyObject *obRet = MultiParmToObject(data);
		data.Free();
		return obRet;
	}

	PyObject * 
	ScriptService<IPropertySrv>::Set(ScriptServiceObject<IPropertySrv> *self, PyObject *args)
	{
		const char *name, *field = NULL;
		object obj;
		cMultiParm data;
		if (!PyArg_ParseTuple(args, "O&szO&:Set", cvtObject, &obj, &name, &field, cvtMultiParm, &data))
		{
			return NULL;
		}
		long ret;
		if (field)
			ret = self->m_pIFace->Set(obj,name,field,data);
		else
			ret = self->m_pIFace->Set(obj,name,data);
		data.Free();
		return PyInt_FromLong(ret);
	}

#if (_DARKGAME == 2) || (_DARKGAME == 3)
	PyObject * 
	ScriptService<IPropertySrv>::SetLocal(ScriptServiceObject<IPropertySrv> *self, PyObject *args)
	{
		const char *name, *field = NULL;
		object obj;
		cMultiParm data;
		if (!PyArg_ParseTuple(args, "O&szO&:SetLocal", cvtObject, &obj, &name, &field, cvtMultiParm, &data))
		{
			return NULL;
		}
		long ret = self->m_pIFace->SetLocal(obj,name,field,data);
		data.Free();
		return PyInt_FromLong(ret);
	}
#endif

	PyObject * 
	ScriptService<IPropertySrv>::Add(ScriptServiceObject<IPropertySrv> *self, PyObject *args)
	{
		const char *name;
		object obj;
		if (!PyArg_ParseTuple(args, "O&s:Add", cvtObject, &obj, &name))
		{
			return NULL;
		}
		long ret = self->m_pIFace->Add(obj,name);
		return PyInt_FromLong(ret);
	}

	PyObject * 
	ScriptService<IPropertySrv>::Remove(ScriptServiceObject<IPropertySrv> *self, PyObject *args)
	{
		const char *name;
		object obj;
		if (!PyArg_ParseTuple(args, "O&s:Remove", cvtObject, &obj, &name))
		{
			return NULL;
		}
		long ret = self->m_pIFace->Remove(obj,name);
		return PyInt_FromLong(ret);
	}

	PyObject * 
	ScriptService<IPropertySrv>::CopyFrom(ScriptServiceObject<IPropertySrv> *self, PyObject *args)
	{
		const char *name;
		object obj, src;
		if (!PyArg_ParseTuple(args, "O&sO&:CopyFrom", cvtObject, &obj, &name, cvtObject, &src))
		{
			return NULL;
		}
		long ret = self->m_pIFace->CopyFrom(obj,name,src);
		return PyInt_FromLong(ret);
	}

	PyObject * 
	ScriptService<IPropertySrv>::Possessed(ScriptServiceObject<IPropertySrv> *self, PyObject *args)
	{
		const char *name;
		object obj;
		if (!PyArg_ParseTuple(args, "O&s:Possessed", cvtObject, &obj, &name))
		{
			return NULL;
		}
		PyObject *obRet = (self->m_pIFace->Possessed(obj,name)) ? Py_True : Py_False;
		Py_INCREF(obRet);
		return obRet;
	}


} // namespace PyNumbra


