/******************************************************************************
 *    PyNumbraExceptions.cc
 *
 *    This file is part of PyNumbra
 *    Copyright (C) 2004 Tom N Harris <telliamed@whoopdedo.cjb.net>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <Python.h>
#include "pydefs.h"
#include <new>
#include <lg/scrservices.h>
#include <lg/scrmanagers.h>

#include "PyNumbraExceptions.h"
#include "pyutils.h"

extern IScriptMan *g_pScriptManager;

namespace PyNumbra
{

	PyObject * Exception::m_Exceptions[] = { NULL };

	void Exception::InitModule(PyObject *module)
	{
		if (!m_Exceptions[0])
		{
			if (!(m_Exceptions[DarkEngineError] = PyErr_NewException("PyNumbra.EngineError", PyExc_EnvironmentError, NULL)))
				m_Exceptions[DarkEngineError] = PyExc_RuntimeError;
			if (!(m_Exceptions[NoInterfaceError] = PyErr_NewException("PyNumbra.NoInterfaceError", PyExc_EnvironmentError, NULL)))
				m_Exceptions[NoInterfaceError] = PyExc_RuntimeError;
		}
		PyObject *moddir = PyModule_GetDict(module);
		PyDict_SetItemString(moddir, "EngineError", reinterpret_cast<PyObject*>(m_Exceptions[DarkEngineError]));
		PyDict_SetItemString(moddir, "NoInterfaceError", reinterpret_cast<PyObject*>(m_Exceptions[NoInterfaceError]));
	}

	void Exception::Raise(Exception::eExceptions type, const char *msg)
	{
		PyErr_SetString(m_Exceptions[type], msg);
	}


	static void
	print_error_text(int offset, const char *text)
	{
		char *nl;
		if (offset >= 0) {
			if (offset > 0 && offset == (int)strlen(text))
				offset--;
			for (;;) {
				nl = strchr(text, '\n');
				if (nl == NULL || nl-text >= offset)
					break;
				offset -= (nl+1-text);
				text = nl+1;
			}
			while (*text == ' ' || *text == '\t') {
				text++;
				offset--;
			}
		}
		monoprintf("    %s", text);
		if (*text == '\0' || text[strlen(text)-1] != '\n')
			monoprintf("\n");
		if (offset == -1)
			return;
		monoprintf("    ");
		offset--;
		while (offset > 0) {
			monoprintf(" ");
			offset--;
		}
		monoprintf("^\n");
	}

	static int
	parse_syntax_error(PyObject *err, PyObject **message, const char **filename,
			   int *lineno, int *offset, const char **text)
	{
		long hold;
		PyObject *v;

		/* old style errors */
		if (PyTuple_Check(err))
			return PyArg_ParseTuple(err, "O(ziiz)", message, filename,
							lineno, offset, text);

		/* new style errors.  `err' is an instance */

		if (! (v = PyObject_GetAttrString(err, "msg")))
			goto finally;
		*message = v;

		if (!(v = PyObject_GetAttrString(err, "filename")))
			goto finally;
		if (v == Py_None)
			*filename = NULL;
		else if (! (*filename = PyString_AsString(v)))
			goto finally;

		Py_DECREF(v);
		if (!(v = PyObject_GetAttrString(err, "lineno")))
			goto finally;
		hold = PyInt_AsLong(v);
		Py_DECREF(v);
		v = NULL;
		if (hold < 0 && PyErr_Occurred())
			goto finally;
		*lineno = (int)hold;

		if (!(v = PyObject_GetAttrString(err, "offset")))
			goto finally;
		if (v == Py_None) {
			*offset = -1;
			Py_DECREF(v);
			v = NULL;
		} else {
			hold = PyInt_AsLong(v);
			Py_DECREF(v);
			v = NULL;
			if (hold < 0 && PyErr_Occurred())
				goto finally;
			*offset = (int)hold;
		}

		if (!(v = PyObject_GetAttrString(err, "text")))
			goto finally;
		if (v == Py_None)
			*text = NULL;
		else if (! (*text = PyString_AsString(v)))
			goto finally;
		Py_DECREF(v);
		return 1;

	finally:
		Py_XDECREF(v);
		return 0;
	}

	void MonoException(const char *prefix, bool direct_to_mono)
	{
		PyObject *obEType, *obEValue, *obTB;
		PyObject *obErrorMsg = NULL, *obErrorName = NULL;

		if (PyErr_ExceptionMatches(PyExc_SystemExit))
		{
			// I can't think of anything reasonable to do here. 
			// So just don't call sys.exit, dammit.
			return;
		}

		monoprintf("*** Python exception during %s\n", prefix);

		if (!direct_to_mono)
		{
			PyErr_Print();
			return;
		}

		PyErr_Fetch(&obEType, &obEValue, &obTB);
		PyErr_NormalizeException(&obEType, &obEValue, &obTB);
		if (!obEType)
			return;
		
		if (PyObject_HasAttrString(obEValue, "print_file_and_line"))
		{
			PyObject *msg;
			const char *fname, *text;
			int line, off;
			if (parse_syntax_error(obEValue, &msg, &fname, &line, &off, &text))
			{
				monoprintf("  File \"%s\", line %d\n", fname ? : "<string>", line);
				if (text)
					print_error_text(off, text);
				obEValue = msg;
			}
		}
		const char *szErrorName = "<exception>";
		obErrorName = PyObject_Str(obEType);
		if (obErrorName)
		{
			szErrorName = PyString_AsString(obErrorName);
			if (!strncmp(szErrorName, "exceptions.", 11))
				szErrorName += 11;
		}

		const char *szErrorMsg = "";
		if (obEValue)
		{
			obErrorMsg = PyObject_Str(obEValue);
			if (obErrorMsg)
				szErrorMsg = PyString_AsString(obErrorMsg);
		}

		monoprintf("*** %s: %s\n", szErrorName, szErrorMsg);

		Py_XDECREF(obErrorMsg);
		Py_XDECREF(obErrorName);
		Py_XDECREF(obTB);
		Py_XDECREF(obEValue);
		Py_XDECREF(obEType);
	}

}
