/******************************************************************************
 *    PyNumbraTypes.cc
 *
 *    This file is part of PyNumbra
 *    Copyright (C) 2004 Tom N Harris <telliamed@whoopdedo.cjb.net>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <Python.h>
#include "pydefs.h"
#include <lg/links.h>
#include <lg/objects.h>
#include <lg/scrmanagers.h>
#include "interface.h"

#include "pyutils.h"
#include "PyNumbraExceptions.h"
#include "PyNumbraModule.h"
#include "PyNumbraTypes.h"

namespace PyNumbra
{

	const char * ObjectObject::Name = "PyNumbra.Object";

	const char * ObjectObject::DocStrings[] = {
		"Object(id_or_name)\n\t"
			"Returns an object representing an object. Err... that is, given the\n\t"
			"object name or object ID of a game object, you get a Python object\n\t"
			"that describes the game object. Yeah, it confuses me too.",
		"ID() -> Integer\n\t"
			"Returns the object ID of this object.",
		"Name() -> String\n\t"
			"Returns the symbolic name of this object.",
		"Exists() -> Boolean\n\t"
			"Check to make sure this object actually exists. In the game,\n\t"
			"I mean. It obviously exists in Python.",
		"Clone(obj)\n\t"
			"Copy the properties from another object onto this one.",
		"GetArchetype() -> Object\n\t"
			"Returns the immediate ancestor of this object.",
		"SetArchetype(obj)\n\t"
			"Makes this object a child of obj.",
		"IsArchetype() -> Boolean\n\t"
			"Returns whether this is an abstract object that can be\n\t"
			"inherited from.",
		"IsMetaProperty() -> Boolean\n\t"
			"Returns whether this is an abstract object that inherits\n\t"
			"from the root meta-property; and thus, can be used as\n\t"
			"a meta-property itself.",
	};

	const PyNumberMethods ObjectObject::NumMethods = {
		0,0,0,0,0,0,0,0,0,0,
		reinterpret_cast<inquiry>(NonZero),
		0,0,0,0,0,0,0,
		reinterpret_cast<unaryfunc>(AsInteger),
		0,0,	/* nb_long, nb_float */
		reinterpret_cast<unaryfunc>(AsOctal),
		reinterpret_cast<unaryfunc>(AsHex),
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	};

	const PyMethodDef ObjectObject::Methods[] = {
		{ "ID", reinterpret_cast<PyCFunction>(GetID), METH_NOARGS, const_cast<char*>(DocStrings[1]) },
		{ "Name", reinterpret_cast<PyCFunction>(GetName), METH_NOARGS, const_cast<char*>(DocStrings[2]) },
		{ "Exists", reinterpret_cast<PyCFunction>(Exists), METH_NOARGS, const_cast<char*>(DocStrings[3]) },
		{ "Clone", reinterpret_cast<PyCFunction>(Clone), METH_O, const_cast<char*>(DocStrings[4]) },
		{ "GetArchetype", reinterpret_cast<PyCFunction>(GetArchetype), METH_NOARGS, const_cast<char*>(DocStrings[5]) },
		{ "SetArchetype", reinterpret_cast<PyCFunction>(SetArchetype), METH_O, const_cast<char*>(DocStrings[6]) },
		{ "IsArchetype", reinterpret_cast<PyCFunction>(IsArchetype), METH_NOARGS, const_cast<char*>(DocStrings[7]) },
		{ "IsMetaProperty", reinterpret_cast<PyCFunction>(IsMetaProperty), METH_NOARGS, const_cast<char*>(DocStrings[8]) },
		{ NULL, NULL, 0, NULL }
	};

	PyTypeObject ObjectObject::Type = {
		PyObject_HEAD_INIT(NULL)
		0,
		const_cast<char*>(Name),
		sizeof(ObjectObject), 0,
		reinterpret_cast<destructor>(Dealloc),
		0,			/* tp_print */
		0, 0,		/* tp_getattr, tp_setattr */
		0,			/* tp_compare */
		reinterpret_cast<reprfunc>(Repr),
		const_cast<PyNumberMethods*>(&NumMethods),
		0, 0,	/* tp_as_sequence, tp_as_mapping */
		reinterpret_cast<hashfunc>(Hash),
		0,		/* tp_call */
		reinterpret_cast<reprfunc>(Str),
		0, 0, 0,	/* tp_getattro, tp_setattro, tp_as_buffer */
		Py_TPFLAGS_DEFAULT,		/* tp_flags */
		const_cast<char*>(DocStrings[0]),
		0, 0,	/* tp_traverse, tp_clear */
		reinterpret_cast<richcmpfunc>(RichCompare),
		0, 0, 0,	/* tp_weaklistoffset, tp_iter, tp_iternext */
		const_cast<PyMethodDef*>(Methods),
		0,0,0,0,0,0,0,
		0,0,
		reinterpret_cast<newfunc>(Construct),
		0,
		0,0,0,0,0,0,0
	};
	

	const char * LinkObject::Name = "PyNumbra.Link";

	const char * LinkObject::DocStrings[] = {
		"Link(id)\n\t"
			"Returns an object representing a link.",
		"ID() -> Integer\n\t"
			"Returns the ID of this link.",
		"To() -> Object\n\t"
			"Returns the destination of this link.",
		"From() -> Object\n\t"
			"Returns the source of this link.",
		"Flavor() -> Integer\n\t"
			"Returns the link flavor number.",
		"GetData()\n\t"
			"Returns the data associated with this link. It may be an Integer,\n\t"
			"Float, String, Boolean, Vector, or Dictionary.",
	};

	const PyNumberMethods LinkObject::NumMethods = {
		0,0,0,0,0,0,0,0,0,0,
		reinterpret_cast<inquiry>(NonZero),
		0,0,0,0,0,0,0,
		reinterpret_cast<unaryfunc>(AsInteger),
		0,0,	/* nb_long, nb_float */
		reinterpret_cast<unaryfunc>(AsOctal),
		reinterpret_cast<unaryfunc>(AsHex),
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	};

	const PyMethodDef LinkObject::Methods[] = {
		{ "ID", reinterpret_cast<PyCFunction>(ID), METH_NOARGS, const_cast<char*>(DocStrings[1]) },
		{ "To", reinterpret_cast<PyCFunction>(To), METH_NOARGS, const_cast<char*>(DocStrings[2]) },
		{ "From", reinterpret_cast<PyCFunction>(From), METH_NOARGS, const_cast<char*>(DocStrings[3]) },
		{ "Flavor", reinterpret_cast<PyCFunction>(Flavor), METH_NOARGS, const_cast<char*>(DocStrings[4]) },
		{ "GetData", reinterpret_cast<PyCFunction>(GetData), METH_NOARGS, const_cast<char*>(DocStrings[5]) },
		{ NULL, NULL, 0, NULL }
	};

	PyTypeObject LinkObject::Type = {
		PyObject_HEAD_INIT(NULL)
		0,
		const_cast<char*>(Name),
		sizeof(LinkObject), 0,
		reinterpret_cast<destructor>(Dealloc),
		0,			/* tp_print */
		0, 0,		/* tp_getattr, tp_setattr */
		0,			/* tp_compare */
		reinterpret_cast<reprfunc>(Repr),
		const_cast<PyNumberMethods*>(&NumMethods),
		0, 0,	/* tp_as_sequence, tp_as_mapping */
		reinterpret_cast<hashfunc>(Hash),
		0,		/* tp_call */
		reinterpret_cast<reprfunc>(Str),
		0, 0, 0,	/* tp_getattro, tp_setattro, tp_as_buffer */
		Py_TPFLAGS_DEFAULT,		/* tp_flags */
		const_cast<char*>(DocStrings[0]),
		0, 0,	/* tp_traverse, tp_clear */
		reinterpret_cast<richcmpfunc>(RichCompare),
		0, 0, 0,	/* tp_weaklistoffset, tp_iter, tp_iternext */
		const_cast<PyMethodDef*>(Methods),
		0,0,0,0,0,0,0,
		0,0,
		reinterpret_cast<newfunc>(Construct),
		0,
		0,0,0,0,0,0,0
	};
	

	const char * LinkIterObject::Name = "PyNumbra.LinkIterator";

	const char * LinkIterObject::DocStrings[] = {
		"LinkIterator()\n\tA LinkIterator encapsulates the interface for retrieving multiple\n\tlinks.",
		"next() -> Link\n\tReturns the next link in the group. Raises StopIteration if there\n\tare no more links.",
	};

	const PyMethodDef LinkIterObject::Methods[] = {
	//	{ "next", reinterpret_cast<PyCFunction>(Next), METH_NOARGS, const_cast<char*>(DocStrings[1]) },
		{ NULL, NULL, 0, NULL }
	};

	PyTypeObject LinkIterObject::Type = {
		PyObject_HEAD_INIT(NULL)
		0,
		const_cast<char*>(Name),
		sizeof(LinkIterObject), 0,
		reinterpret_cast<destructor>(Dealloc),
		0,			/* tp_print */
		0, 0,		/* tp_getattr, tp_setattr */
		0,			/* tp_compare */
		reinterpret_cast<reprfunc>(Repr),
		0,
		0, 0,	/* tp_as_sequence, tp_as_mapping */
		0,
		0,		/* tp_call */
		0,
		0, 0, 0,	/* tp_getattro, tp_setattro, tp_as_buffer */
		Py_TPFLAGS_DEFAULT,		/* tp_flags */
		const_cast<char*>(DocStrings[0]),
		0, 0,	/* tp_traverse, tp_clear */
		0, // LinkIterators always compare unequal (except for identity)
		0,	/* tp_weaklistoffset */
		reinterpret_cast<getiterfunc>(Iter), 
		reinterpret_cast<iternextfunc>(Next),
		const_cast<PyMethodDef*>(Methods),
		0,0,0,0,0,0,0,
		0,0,
		0, // No default constructor
		0,
		0,0,0,0,0,0,0
	};
	

	const char * TimerObject::Name = "Timer";

	const char * TimerObject::DocStrings[] = {
		"Timer(handle)\n\tReturns the timer with the given integer handle.",
		"kill()\n\tAborts the timer.",
	};

	const PyNumberMethods TimerObject::NumMethods = {
		0,0,0,0,0,0,0,0,0,0,
		reinterpret_cast<inquiry>(NonZero),
		0,0,0,0,0,0,0,
		reinterpret_cast<unaryfunc>(AsInteger),
		0,0,	/* nb_long, nb_float */
		reinterpret_cast<unaryfunc>(AsOctal),
		reinterpret_cast<unaryfunc>(AsHex),
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	};

	const PyMethodDef TimerObject::Methods[] = {
		{ "kill", reinterpret_cast<PyCFunction>(KillTimer), METH_NOARGS, const_cast<char*>(DocStrings[1]) },
		{ NULL, NULL, 0, NULL }
	};

	PyTypeObject TimerObject::Type = {
		PyObject_HEAD_INIT(NULL)
		0,
		const_cast<char*>(Name),
		sizeof(TimerObject), 0,
		reinterpret_cast<destructor>(Dealloc),
		0,			/* tp_print */
		0, 0,		/* tp_getattr, tp_setattr */
		0,			/* tp_compare */
		reinterpret_cast<reprfunc>(Repr),
		const_cast<PyNumberMethods*>(&NumMethods),
		0, 0,	/* tp_as_sequence, tp_as_mapping */
		reinterpret_cast<hashfunc>(Hash),
		0,		/* tp_call */
		0,
		0, 0, 0,	/* tp_getattro, tp_setattro, tp_as_buffer */
		Py_TPFLAGS_DEFAULT,		/* tp_flags */
		const_cast<char*>(DocStrings[0]),
		0, 0,	/* tp_traverse, tp_clear */
		reinterpret_cast<richcmpfunc>(RichCompare),
		0, 0, 0,	/* tp_weaklistoffset, tp_iter, tp_iternext */
		const_cast<PyMethodDef*>(Methods),
		0,0,0,0,0,0,0,
		0,0,
		reinterpret_cast<newfunc>(Construct),
		0,
		0,0,0,0,0,0,0
	};
	

/////////////////
// ObjectObject
/////////////////
	void ObjectObject::InitModule(PyObject *module)
	{
		PyType_Ready(&Type);

		PyDict_SetItemString(PyModule_GetDict(module), "Object", reinterpret_cast<PyObject*>(&Type));

		PyObject *cpmod = PyImport_ImportModule("copy_reg");
		if (cpmod)
		{
			PyObject *p = PyString_FromString("pickle");
			PyObject *cpreg = PyObject_GetAttr(cpmod, p);
			Py_XDECREF(p);
			PyObject *moddir = PyModule_GetDict(module);
			PyObject *confunc = PyDict_GetItemString(moddir, "_construct_object");
			PyObject *cpfunc = PyDict_GetItemString(moddir, "_pickle_object");
			PyObject *result = PyObject_CallFunction(cpreg, "(OOO)", &Type, cpfunc, confunc);
			Py_XDECREF(result);
			Py_DECREF(cpmod);
		}
	}

	PyObject * ObjectObject::New(object o)
	{
		ObjectObject *self = PyObject_New(ObjectObject, &Type);
		if (self)
		{
			self->m_object = o;
			try
			{
				CObjectSystem.Acquire();
				CTraitManager.Acquire();
			}
			catch (no_interface &err)
			{
				Exception::Raise(Exception::NoInterfaceError, "PyNumbra.Object");
				PyObject_Del(self);
				return NULL;
			}
		}
		return static_cast<PyObject*>(self);
	}

	PyObject * ObjectObject::Construct(PyTypeObject *type, PyObject *args, PyObject *)
	{
		int id = 0;
		PyObject *obArg;
		if (!PyArg_ParseTuple(args, "O:Object", &obArg))
		{
			PyErr_SetString(PyExc_TypeError, "argument must be an object ID or name");
			return NULL;
		}
		try 
		{
			CObjectSystem.Acquire();
			CTraitManager.Acquire();
		}
		catch (no_interface &err)
		{
			Exception::Raise(Exception::NoInterfaceError, "PyNumbra.Object");
			return NULL;
		}
		if (PyObject_TypeCheck(obArg, &Type))
		{
			id = static_cast<int>(static_cast<ObjectObject*>(obArg)->m_object);
		}
		else if (PyString_Check(obArg) || PyUnicode_Check(obArg))
		{
			char *buf;
			if (0 > PyString_AsStringAndSize(obArg, &buf, NULL))
			{
				PyErr_SetString(PyExc_ValueError, "object name cannot contain embedded NULL characters");
				CObjectSystem.Release();
				CTraitManager.Release();
				return NULL;
			}
			id = CObjectSystem->GetObjectNamed(buf);
		}
		else
		{
			PyObject *num = PyNumber_Int(obArg);
			if (!num)
			{
				PyErr_SetString(PyExc_TypeError, "argument must be an object ID or name");
				CObjectSystem.Release();
				CTraitManager.Release();
				return NULL;
			}
			id = PyInt_AsLong(num);
			Py_DECREF(num);
		}
		ObjectObject *self = (type && PyType_CheckExact(type)) ? static_cast<ObjectObject*>(type->tp_alloc(type, 0))
						: PyObject_New(ObjectObject, &Type);
		if (self)
		{
			self->m_object = id;
		}
		else
		{
			CObjectSystem.Release();
			CTraitManager.Release();
		}
		return static_cast<PyObject*>(self);
	}

	PyObject * ObjectObject::Pickle(PyObject *, ObjectObject *self)
	{
		// There's no need to import, since we know it exists.
		PyObject *mod = PyImport_AddModule("PyNumbra");
		PyObject *confunc = PyDict_GetItemString(PyModule_GetDict(mod), "_construct_object");
		PyObject *obRet = Py_BuildValue("(O(i))", confunc, static_cast<int>(self->m_object));
		return obRet;
	}

	void ObjectObject::Dealloc(ObjectObject *self)
	{
		CObjectSystem.Release();
		CTraitManager.Release();
		self->ob_type->tp_free(static_cast<PyObject*>(self));
	}

	PyObject * ObjectObject::Repr(ObjectObject *self)
	{
		return PyString_FromFormat("<PyNumbra Object at %p (%d)>", self, static_cast<int>(self->m_object));
	}

	PyObject * ObjectObject::Str(ObjectObject *self)
	{
		const char *name = NULL;
		if (self->m_object != 0)
			name = CObjectSystem->GetName(self->m_object);
		PyObject *obRet;
		if (name)
			obRet = PyString_FromFormat("<%s (%d)>", name, static_cast<int>(self->m_object));
		else
			obRet = PyString_FromFormat("<Object (%d)>", static_cast<int>(self->m_object));
		return obRet;
	}

	PyObject * ObjectObject::AsInteger(ObjectObject *self)
	{
		return PyInt_FromLong(self->m_object);
	}

	PyObject * ObjectObject::AsOctal(ObjectObject *self)
	{
		char buf[12];
		sprintf(buf, "0%o", static_cast<int>(self->m_object));
		return PyString_FromString(buf);
	}

	PyObject * ObjectObject::AsHex(ObjectObject *self)
	{
		char buf[10];
		sprintf(buf, "0x%x", static_cast<int>(self->m_object));
		return PyString_FromString(buf);
	}

	int ObjectObject::NonZero(ObjectObject *self)
	{
		return self->m_object != 0;
	}

	PyObject * ObjectObject::RichCompare(PyObject *self, PyObject *other, int op)
	{
		int a,b;
		if (PyObject_TypeCheck(self, &Type))
			a = static_cast<int>(static_cast<ObjectObject*>(self)->m_object);
		else
		{
			PyObject *num = PyNumber_Int(self);
			if (!num)
			{
				Py_INCREF(Py_NotImplemented);
				return Py_NotImplemented;
			}
			a = PyInt_AsLong(num);
			Py_DECREF(num);
		}

		if (PyObject_TypeCheck(other, &Type))
			b = static_cast<int>(static_cast<ObjectObject*>(other)->m_object);
		else
		{
			PyObject *num = PyNumber_Int(other);
			if (!num)
			{
				Py_INCREF(Py_NotImplemented);
				return Py_NotImplemented;
			}
			b = PyInt_AsLong(num);
			Py_DECREF(num);
		}

		PyObject *obRet = Py_False;
		switch (op)
		{
			case Py_EQ:
				if (a == b) obRet = Py_True;
				break;
			case Py_NE:
				if (a != b) obRet = Py_True;
				break;
			case Py_LT:
				if (a < b) obRet = Py_True;
				break;
			case Py_LE:
				if (a <= b) obRet = Py_True;
				break;
			case Py_GT:
				if (a > b) obRet = Py_True;
				break;
			case Py_GE:
				if (a >= b) obRet = Py_True;
				break;
		}
		Py_INCREF(obRet);
		return obRet;
	}

	long ObjectObject::Hash(ObjectObject *self)
	{
		return self->m_object;
	}

	PyObject * ObjectObject::GetID(ObjectObject *self)
	{
		return PyInt_FromLong(self->m_object);
	}

	PyObject * ObjectObject::GetName(ObjectObject *self)
	{
		const char *name = CObjectSystem->GetName(self->m_object);
		if (!name)
		{
			Py_INCREF(Py_None);
			return Py_None;
		}
		return PyString_FromString(name);
	}

	PyObject * ObjectObject::Exists(ObjectObject *self)
	{
		PyObject *obRet = (CObjectSystem->Exists(self->m_object)) ? Py_True : Py_False;
		Py_INCREF(obRet);
		return obRet;
	}

	PyObject * ObjectObject::Clone(ObjectObject *self, PyObject *arg)
	{
		PyObject *obId = PyNumber_Int(arg);
		if (!obId)
		{
			return NULL;
		}
		int id = PyInt_AsLong(obId);
		Py_DECREF(obId);
		if (id == 0 || self->m_object == 0)
		{
			PyErr_SetString(PyExc_ValueError, "object ID must be non-zero");
			return NULL;
		}
		long ret = CObjectSystem->CloneObject(self->m_object, id);
		return PyInt_FromLong(ret);
	}

	PyObject * ObjectObject::GetArchetype(ObjectObject *self)
	{
		return New(CTraitManager->GetArchetype(self->m_object));
	}

	PyObject * ObjectObject::SetArchetype(ObjectObject *self, PyObject *arg)
	{
		PyObject *obId = PyNumber_Int(arg);
		if (!obId)
		{
			return NULL;
		}
		int id = PyInt_AsLong(obId);
		Py_DECREF(obId);
		if (id == 0 || self->m_object == 0)
		{
			PyErr_SetString(PyExc_ValueError, "archetype ID must be non-zero");
			return NULL;
		}
		long ret = CTraitManager->SetArchetype(self->m_object, id);
		return PyInt_FromLong(ret);
	}

	PyObject * ObjectObject::IsArchetype(ObjectObject *self)
	{
		PyObject *obRet = (CTraitManager->IsArchetype(self->m_object)) ? Py_True : Py_False;
		Py_INCREF(obRet);
		return obRet;
	}

	PyObject * ObjectObject::IsMetaProperty(ObjectObject *self)
	{
		PyObject *obRet = (CTraitManager->IsMetaProperty(self->m_object)) ? Py_True : Py_False;
		Py_INCREF(obRet);
		return obRet;
	}


///////////////
// LinkObject
///////////////
	void LinkObject::InitModule(PyObject *module)
	{
		PyType_Ready(&Type);

		PyDict_SetItemString(PyModule_GetDict(module), "Link", reinterpret_cast<PyObject*>(&Type));

		PyObject *cpmod = PyImport_ImportModule("copy_reg");
		if (cpmod)
		{
			PyObject *p = PyString_FromString("pickle");
			PyObject *cpreg = PyObject_GetAttr(cpmod, p);
			Py_XDECREF(p);
			PyObject *moddir = PyModule_GetDict(module);
			PyObject *confunc = PyDict_GetItemString(moddir, "_construct_object");
			PyObject *cpfunc = PyDict_GetItemString(moddir, "_pickle_object");
			PyObject *result = PyObject_CallFunction(cpreg, "(OOO)", &Type, cpfunc, confunc);
			Py_XDECREF(result);
			Py_DECREF(cpmod);
		}
	}

	PyObject * LinkObject::New(link l)
	{
		LinkObject *self = PyObject_New(LinkObject, &Type);
		if (self)
		{
			self->m_link = l;
			self->m_flavor = 0;
			try 
			{
				CLinkManager.Acquire();
			}
			catch (no_interface &err)
			{
				Exception::Raise(Exception::NoInterfaceError, "PyNumbra.Link");
				PyObject_Del(self);
				return NULL;
			}
		}
		return static_cast<PyObject*>(self);
	}

	PyObject * LinkObject::Construct(PyTypeObject *type, PyObject *args, PyObject *)
	{
		long id = 0;
		PyObject *obArg;
		if (!PyArg_ParseTuple(args, "O:Link", &obArg))
		{
			PyErr_SetString(PyExc_TypeError, "argument must be a link ID");
			return NULL;
		}
		if (PyObject_TypeCheck(obArg, &Type))
		{
			id = static_cast<long>(static_cast<LinkObject*>(obArg)->m_link);
		}
		else
		{
			PyObject *num = PyNumber_Int(obArg);
			if (!num)
			{
				PyErr_SetString(PyExc_TypeError, "argument must be a link ID");
				return NULL;
			}
			id = PyInt_AsLong(num);
			Py_DECREF(num);
		}
		LinkObject *self = (type && PyType_CheckExact(type)) ? static_cast<LinkObject*>(type->tp_alloc(type, 0))
						: PyObject_New(LinkObject, &Type);
		if (self)
		{
			self->m_link = id;
			self->m_flavor = 0;
			try 
			{
				CLinkManager.Acquire();
			}
			catch (no_interface &err)
			{
				Exception::Raise(Exception::NoInterfaceError, "PyNumbra.Link");
				PyObject_Del(self);
				return NULL;
			}
		}
		return static_cast<PyObject*>(self);
	}

	PyObject * LinkObject::Pickle(PyObject *, LinkObject *self)
	{
		// There's no need to import, since we know it exists.
		PyObject *mod = PyImport_AddModule("PyNumbra");
		PyObject *confunc = PyDict_GetItemString(PyModule_GetDict(mod), "_construct_link");
		PyObject *obRet = Py_BuildValue("(O(l))", confunc, static_cast<long>(self->m_link));
		return obRet;
	}

	void LinkObject::Dealloc(LinkObject *self)
	{
		CLinkManager.Release();
		self->ob_type->tp_free(static_cast<PyObject*>(self));
	}

	PyObject * LinkObject::Repr(LinkObject *self)
	{
		return PyString_FromFormat("<PyNumbra Link at %p (%x)>", self, static_cast<unsigned int>(self->m_link));
	}

	PyObject * LinkObject::Str(LinkObject *self)
	{
		return PyString_FromFormat("<Link %x>", static_cast<unsigned int>(self->m_link));
	}

	PyObject * LinkObject::AsInteger(LinkObject *self)
	{
		return PyInt_FromLong(self->m_link);
	}

	PyObject * LinkObject::AsOctal(LinkObject *self)
	{
		char buf[12];
		sprintf(buf, "0%o", static_cast<long>(self->m_link));
		return PyString_FromString(buf);
	}

	PyObject * LinkObject::AsHex(LinkObject *self)
	{
		char buf[10];
		sprintf(buf, "0x%x", static_cast<long>(self->m_link));
		return PyString_FromString(buf);
	}

	int LinkObject::NonZero(LinkObject *self)
	{
		return self->m_link != 0;
	}

	PyObject * LinkObject::RichCompare(PyObject *self, PyObject *other, int op)
	{
		long a,b;
		if (PyObject_TypeCheck(self, &Type))
			a = static_cast<long>(static_cast<LinkObject*>(self)->m_link);
		else
		{
			PyObject *num = PyNumber_Int(self);
			if (!num)
			{
				Py_INCREF(Py_NotImplemented);
				return Py_NotImplemented;
			}
			a = PyInt_AsLong(num);
			Py_DECREF(num);
		}

		if (PyObject_TypeCheck(other, &Type))
			b = static_cast<long>(static_cast<LinkObject*>(other)->m_link);
		else
		{
			PyObject *num = PyNumber_Int(other);
			if (!num)
			{
				Py_INCREF(Py_NotImplemented);
				return Py_NotImplemented;
			}
			b = PyInt_AsLong(num);
			Py_DECREF(num);
		}

		PyObject *obRet = Py_False;
		switch (op)
		{
			case Py_EQ:
				if (a == b) obRet = Py_True;
				break;
			case Py_NE:
				if (a != b) obRet = Py_True;
				break;
			case Py_LT:
				if (a < b) obRet = Py_True;
				break;
			case Py_LE:
				if (a <= b) obRet = Py_True;
				break;
			case Py_GT:
				if (a > b) obRet = Py_True;
				break;
			case Py_GE:
				if (a >= b) obRet = Py_True;
				break;
		}
		Py_INCREF(obRet);
		return obRet;
	}

	long LinkObject::Hash(LinkObject *self)
	{
		return self->m_link;
	}

	PyObject * LinkObject::ID(LinkObject *self)
	{
		return PyInt_FromLong(self->m_link);
	}

	PyObject * LinkObject::To(LinkObject *self)
	{
		if (self->m_link == 0)
		{
			Py_INCREF(Py_None);
			return Py_None;
		}
		if (self->m_flavor == 0)
		{
			sLink info;
			CLinkManager->Get(self->m_link, &info);
			self->m_source = info.source;
			self->m_dest = info.dest;
			self->m_flavor = info.flavor;
		}
		return ObjectObject::New(self->m_dest);
	}

	PyObject * LinkObject::From(LinkObject *self)
	{
		if (self->m_link == 0)
		{
			Py_INCREF(Py_None);
			return Py_None;
		}
		if (self->m_flavor == 0)
		{
			sLink info;
			CLinkManager->Get(self->m_link, &info);
			self->m_source = info.source;
			self->m_dest = info.dest;
			self->m_flavor = info.flavor;
		}
		return ObjectObject::New(self->m_source);
	}

	PyObject * LinkObject::Flavor(LinkObject *self)
	{
		if (self->m_link == 0)
		{
			Py_INCREF(Py_None);
			return Py_None;
		}
		if (self->m_flavor == 0)
		{
			sLink info;
			CLinkManager->Get(self->m_link, &info);
			self->m_source = info.source;
			self->m_dest = info.dest;
			self->m_flavor = info.flavor;
		}
		return PyInt_FromLong(self->m_flavor);
	}

	PyObject * LinkObject::GetData(LinkObject *self)
	{
		if (self->m_link == 0)
		{
			Py_INCREF(Py_None);
			return Py_None;
		}
		if (self->m_flavor == 0)
		{
			sLink info;
			CLinkManager->Get(self->m_link, &info);
			self->m_source = info.source;
			self->m_dest = info.dest;
			self->m_flavor = info.flavor;
		}
		IRelation *pRel = CLinkManager->GetRelation(self->m_flavor);
		if (!pRel)
		{
			Exception::Raise(Exception::NoInterfaceError, "GetRelation(flavor)");
			return NULL;
		}
		const sRelationDataDesc *pDataDesc = pRel->DescribeData();
		const char *pData = reinterpret_cast<const char *>(pRel->GetData(self->m_link));
		if (!pDataDesc || pDataDesc->uiTypeSize == 0 || !pData)
		{
			pRel->Release();
			Py_INCREF(Py_None);
			return Py_None;
		}
		PyObject *obRet = StructDataToObject(pDataDesc->szTypeName, pData);
		pRel->Release();
		return obRet;
	}


/////////////////
// LinkIterator
/////////////////
	void LinkIterObject::InitModule(PyObject *module)
	{
		PyType_Ready(&Type);
		PyDict_SetItemString(PyModule_GetDict(module), "LinkIterator", reinterpret_cast<PyObject*>(&Type));
	}

	PyObject * LinkIterObject::New(ILinkQuery *q)
	{
		LinkIterObject *self = PyObject_New(LinkIterObject, &Type);
		if (self)
		{
			self->m_query = q;
		}
		return static_cast<PyObject*>(self);
	}

	void LinkIterObject::Dealloc(LinkIterObject *self)
	{
		if (self->m_query)
			self->m_query->Release();
		self->ob_type->tp_free(static_cast<PyObject*>(self));
	}

	PyObject * LinkIterObject::Repr(LinkIterObject *self)
	{
		return PyString_FromFormat("<LinkIterator at %p>", self);
	}

	PyObject * LinkIterObject::Iter(LinkIterObject *self)
	{
		Py_INCREF(self);
		return static_cast<PyObject*>(self);
	}

	PyObject * LinkIterObject::Next(LinkIterObject *self)
	{
		PyObject *obRet = NULL;
		if (self->m_query && !self->m_query->Done())
		{
			obRet = LinkObject::New(self->m_query->ID());
			self->m_query->Next();
		}
		return obRet;
	}


////////////////
// TimerObject
////////////////
	void TimerObject::InitModule(PyObject *module)
	{
		PyType_Ready(&Type);

		PyDict_SetItemString(PyModule_GetDict(module), "Timer", reinterpret_cast<PyObject*>(&Type));

		PyObject *cpmod = PyImport_ImportModule("copy_reg");
		if (cpmod)
		{
			PyObject *p = PyString_FromString("pickle");
			PyObject *cpreg = PyObject_GetAttr(cpmod, p);
			Py_XDECREF(p);
			PyObject *moddir = PyModule_GetDict(module);
			PyObject *confunc = PyDict_GetItemString(moddir, "_construct_object");
			PyObject *cpfunc = PyDict_GetItemString(moddir, "_pickle_object");
			PyObject *result = PyObject_CallFunction(cpreg, "(OOO)", &Type, cpfunc, confunc);
			Py_XDECREF(result);
			Py_DECREF(cpmod);
		}
	}

	PyObject * TimerObject::New(tScrTimer h)
	{
		TimerObject *self = PyObject_New(TimerObject, &Type);
		if (self)
		{
			self->m_handle = h;
		}
		return static_cast<PyObject*>(self);
	}

	PyObject * TimerObject::Construct(PyTypeObject *type, PyObject *args, PyObject *)
	{
		tScrTimer handle = NULL;
		PyObject *obArg;
		if (!PyArg_ParseTuple(args, "O:Timer", &obArg))
		{
			PyErr_SetString(PyExc_TypeError, "argument must be a timer handle");
			return NULL;
		}
		if (PyObject_TypeCheck(obArg, &Type))
		{
			handle = static_cast<TimerObject*>(obArg)->m_handle;
		}
		else
		{
			PyObject *num = PyNumber_Int(obArg);
			if (!num)
			{
				PyErr_SetString(PyExc_TypeError, "argument must be a timer handle");
				return NULL;
			}
			handle = reinterpret_cast<tScrTimer>(PyInt_AsLong(num));
			Py_DECREF(num);
		}
		TimerObject *self = (type && PyType_CheckExact(type)) ? static_cast<TimerObject*>(type->tp_alloc(type, 0))
						: PyObject_New(TimerObject, &Type);
		if (self)
		{
			self->m_handle = handle;
		}
		return static_cast<PyObject*>(self);
	}

	PyObject * TimerObject::Pickle(PyObject *, TimerObject *self)
	{
		// There's no need to import, since we know it exists.
		PyObject *mod = PyImport_AddModule("PyNumbra");
		PyObject *confunc = PyDict_GetItemString(PyModule_GetDict(mod), "_construct_timer");
		PyObject *obRet = Py_BuildValue("(O(l))", confunc, reinterpret_cast<long>(self->m_handle));
		return obRet;
	}

	void TimerObject::Dealloc(TimerObject *self)
	{
		self->ob_type->tp_free(static_cast<PyObject*>(self));
	}

	PyObject * TimerObject::Repr(TimerObject *self)
	{
		return PyString_FromFormat("<Timer at %p (%x)>", self, reinterpret_cast<unsigned int>(self->m_handle));
	}

	PyObject * TimerObject::AsInteger(TimerObject *self)
	{
		return PyInt_FromLong(reinterpret_cast<long>(self->m_handle));
	}

	PyObject * TimerObject::AsOctal(TimerObject *self)
	{
		char buf[12];
		sprintf(buf, "0%o", reinterpret_cast<long>(self->m_handle));
		return PyString_FromString(buf);
	}

	PyObject * TimerObject::AsHex(TimerObject *self)
	{
		char buf[10];
		sprintf(buf, "0x%x", reinterpret_cast<long>(self->m_handle));
		return PyString_FromString(buf);
	}

	int TimerObject::NonZero(TimerObject *self)
	{
		return self->m_handle != NULL;
	}

	PyObject * TimerObject::RichCompare(PyObject *self, PyObject *other, int op)
	{
		tScrTimer a,b;
		if (PyObject_TypeCheck(self, &Type))
			a = static_cast<TimerObject*>(self)->m_handle;
		else
		{
			PyObject *num = PyNumber_Int(self);
			if (!num)
			{
				Py_INCREF(Py_NotImplemented);
				return Py_NotImplemented;
			}
			a = reinterpret_cast<tScrTimer>(PyInt_AsLong(num));
			Py_DECREF(num);
		}

		if (PyObject_TypeCheck(other, &Type))
			b = static_cast<TimerObject*>(other)->m_handle;
		else
		{
			PyObject *num = PyNumber_Int(other);
			if (!num)
			{
				Py_INCREF(Py_NotImplemented);
				return Py_NotImplemented;
			}
			b = reinterpret_cast<tScrTimer>(PyInt_AsLong(num));
			Py_DECREF(num);
		}

		PyObject *obRet = Py_False;
		switch (op)
		{
			case Py_EQ:
				if (a == b) obRet = Py_True;
				break;
			case Py_NE:
				if (a != b) obRet = Py_True;
				break;
			case Py_LT:
				if (a < b) obRet = Py_True;
				break;
			case Py_LE:
				if (a <= b) obRet = Py_True;
				break;
			case Py_GT:
				if (a > b) obRet = Py_True;
				break;
			case Py_GE:
				if (a >= b) obRet = Py_True;
				break;
		}
		Py_INCREF(obRet);
		return obRet;
	}

	long TimerObject::Hash(TimerObject *self)
	{
		return reinterpret_cast<long>(self->m_handle);
	}

	PyObject * TimerObject::KillTimer(TimerObject *self)
	{
		g_pScriptManager->KillTimedMessage(self->m_handle);
		Py_INCREF(Py_None);
		return Py_None;
	}


} // namespace PyNumbra
