/******************************************************************************
 *    ScriptMessages.cc
 *
 *    This file is part of PyNumbra
 *    Copyright (C) 2004 Tom N Harris <telliamed@whoopdedo.cjb.net>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <Python.h>
#include "pydefs.h"
#include <exception>
#include <algorithm>
#include <cstring>

#include <lg/types.h>
#include <lg/scrmsgs.h>

#include "pyutils.h"
#include "PyNumbraTypes.h"
#include "ScriptMessages.h"

namespace PyNumbra
{

	class value_error : public std::exception
	{
	public:
		value_error() throw() { }
		virtual ~value_error() throw() { }

		virtual const char* what() const throw()
		{ return "invalid argument"; }
	};


	PyObject * ScriptMessages::MessageToObject(const sScrMsg *msg)
	{
		PyObject *msgDict = PyDict_New();
		if (!msgDict)
			return NULL;
		PyObject *obElem = PyString_FromString(msg->message);
		PyDict_SetItemString(msgDict, "message", obElem);
		Py_DECREF(obElem);
		obElem = ObjectObject::New(msg->from);
		PyDict_SetItemString(msgDict, "from", obElem);
		Py_DECREF(obElem);
		obElem = ObjectObject::New(msg->to);
		PyDict_SetItemString(msgDict, "to", obElem);
		Py_DECREF(obElem);
		obElem = PyInt_FromLong(msg->time);
		PyDict_SetItemString(msgDict, "time", obElem);
		Py_DECREF(obElem);
		PyObject *obData = PyTuple_New(3);
		obElem = MultiParmToObject(msg->data);
		PyTuple_SET_ITEM(obData, 0, obElem);
		obElem = MultiParmToObject(msg->data2);
		PyTuple_SET_ITEM(obData, 1, obElem);
		obElem = MultiParmToObject(msg->data3);
		PyTuple_SET_ITEM(obData, 2, obElem);
		PyDict_SetItemString(msgDict, "data", obData);
		Py_DECREF(obData);

		_NewObj(msg, msgDict);

		return msgDict;
	}

	sScrMsg * ScriptMessages::MessageFromObject(PyObject *msgDict)
	{
		sScrMsg *ret = NULL;
		cMultiParm data[3];
		PyObject *obName = NULL, *obSource = NULL, *obDest = NULL, *obData = NULL;
		try
		{
			if (!(obName = PyDict_GetItemString(msgDict, "message")))
				throw value_error();
			Py_INCREF(obName);
			if (!PyObject_TypeCheck(obName, &PyBaseString_Type) || PyObject_Length(obName) <= 0)
				throw value_error();
			if (!(obSource = PyDict_GetItemString(msgDict, "from")))
				throw value_error();
			Py_INCREF(obSource);
			if (!(PyInt_Check(obSource) || PyLong_Check(obSource) || PyObject_TypeCheck(obSource, &ObjectObject::Type)))
				throw value_error();
			Py_DECREF(obSource);
			if (!(obSource = PyNumber_Int(obSource)))
				throw value_error();
			if (!(obDest = PyDict_GetItemString(msgDict, "to")))
				throw value_error();
			Py_INCREF(obDest);
			if (!(PyInt_Check(obDest) || PyLong_Check(obDest) || PyObject_TypeCheck(obDest, &ObjectObject::Type)))
				throw value_error();
			Py_DECREF(obDest);
			if (!(obDest = PyNumber_Int(obDest)))
				throw value_error();

			obData = PyDict_GetItemString(msgDict, "data");
			if (obData)
			{
				if (PySequence_Check(obData) && PySequence_Size(obData) <= 3)
				{
					for (int n = 0; n < PySequence_Size(obData); ++n)
					{
						PyObject *elem = PySequence_GetItem(obData, n);
						if (!MultiParmFromObject(data[n], elem))
						{
							while (n--)
							{
								data[n].Free();
							}
							throw value_error();
						}
					}
				}
				else
					throw value_error();
			}

			if (!PyString_CheckExact(obName))
			{
				PyObject *str = PyObject_Str(obName);
				Py_DECREF(obName);
				obName = str;
			}
			ret = _NewMsg(PyString_AsString(obName), msgDict);
			if (!ret)
			{
				ret = new sScrMsg();
				// The defined msg functions should set this to 
				// a static string so we don't have to intern.
				PyString_InternInPlace(&obName);
				// Python 2.2 never destroys interned strings. 
				// But 2.3 will when the only reference to the object
				// is the intern table. We want the string to remain 
				// valid for a long time, so we INCREF it.
				// The memory leakage shouldn't be a problem unless
				// you are in the habit of sending large numbers of 
				// messages each with a different name.
				Py_INCREF(obName);
				// Alternately, I could rewrite sScrMsg to free the 
				// message names, but that's a lot of work when 99%
				// of messages are constructed using static strings.
				ret->message = PyString_AsString(obName);
			}
			else if (!ret->message)
			{
				// this will happen if it's a Stimulus message
				PyString_InternInPlace(&obName);
				ret->message = PyString_AsString(obName);
			}
			ret->from = PyInt_AsLong(obSource);
			ret->to = PyInt_AsLong(obDest);

			ret->data = data[0];
			ret->data2 = data[1];
			ret->data3 = data[2];
		}
		catch (value_error &err)
		{
			PyErr_SetString(PyExc_ValueError, "argument is not a valid message struct");
		}
		catch (std::exception &err)
		{
			PyErr_SetString(PyExc_EnvironmentError, err.what());
		}

		//Py_XDECREF(obData);
		Py_XDECREF(obDest);
		Py_XDECREF(obSource);
		Py_XDECREF(obName);
		return ret;
	}

	inline bool ScriptMessages::_MsgTypeStruct::operator== (const char* pszKey) const 
	{
		return 0 == stricmp(Name, pszKey);
	}

	sScrMsg * ScriptMessages::_NewMsg(const char *pszName, PyObject *msgDict)
	{
		sScrMsg *ret = NULL;
		const _MsgTypeStruct *def = std::find(_MsgTypes, _MsgLast, pszName);
		if (def != _MsgLast)
		{
			ret = def->NewMsg(msgDict);
			ret->message = def->Name;
		}
		else
		{
			int nameLength = strlen(pszName) - 8;
			if (nameLength > 0)
			{
				if (!stricmp(pszName + nameLength, "Stimulus"))
				{
					ret = NewStimMsg(msgDict);
					ret->message = NULL;
				}
			}
		}
		return ret;
	}

	void ScriptMessages::_NewObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const _MsgTypeStruct *def = std::find(_MsgTypes, _MsgLast, msg->message);
		if (def != _MsgLast)
		{
			def->NewObj(msg, msgDict);
		}
#ifdef _MSC_VER
		else if (dynamic_cast<sPersistent*>(msg) != NULL
			 && !stricmp(msg->GetName(), "sStimMsg"))
#else
		// GCC & Borland need to use an assembly kludge, so we 
		// check for a valid vtable there.
		else if (!stricmp(msg->Persistent_GetName(), "sStimMsg"))
#endif
		{
			NewStimObj(msg, msgDict);
		}
	}

	// Keep this list sorted (case-insensitive). We do a binary search.
	const ScriptMessages::_MsgTypeStruct
	ScriptMessages::_MsgTypes[] = {
		{ "AIModeChange", NewAIModeChangeMsg, NewAIModeChangeObj },
		{ "Alertness", NewAlertnessMsg, NewAlertnessObj },
		{ "Call", NewMovingTerrainMsg, NewMovingTerrainObj },
		{ "Combine", NewCombineMsg, NewCombineObj },
		{ "Contained", NewContainedMsg, NewContainedObj },
		{ "Container", NewContainerMsg, NewContainerObj },
		{ "Create", NewGenericMsg, NewGenericObj },
		{ "CreatureRoomEnter", NewRoomMsg, NewRoomObj },
		{ "CreatureRoomExit", NewRoomMsg, NewRoomObj },
		{ "Damage", NewDamageMsg, NewDamageObj },
		{ "DarkGameModeChange", NewSimMsg, NewSimObj },
		// Sent by VK's DarkHooks.dll
		{ "DHGameModeChange", NewSimMsg, NewSimObj },
		{ "Difficulty", NewDifficultyMsg, NewDifficultyObj },
		{ "DoorClose", NewDoorMsg, NewDoorObj },
		{ "DoorClosed", NewDoorMsg, NewDoorObj },
		{ "DoorClosing", NewDoorMsg, NewDoorObj },
		{ "DoorHalt", NewDoorMsg, NewDoorObj },
		{ "DoorOpen", NewDoorMsg, NewDoorObj },
		{ "DoorOpening", NewDoorMsg, NewDoorObj },
		{ "EndAttack", NewAttackMsg, NewAttackObj },
		{ "FrobInvBegin", NewFrobMsg, NewFrobObj },
		{ "FrobInvEnd", NewFrobMsg, NewFrobObj },
		{ "FrobToolBegin", NewFrobMsg, NewFrobObj },
		{ "FrobToolEnd", NewFrobMsg, NewFrobObj },
		{ "FrobWorldBegin", NewFrobMsg, NewFrobObj },
		{ "FrobWorldEnd", NewFrobMsg, NewFrobObj },
		{ "HighAlert", NewHighAlertMsg, NewHighAlertObj },
		{ "InvDeFocus", NewGenericMsg, NewGenericObj },
		{ "InvDeSelect", NewGenericMsg, NewGenericObj },
		{ "InvFocus", NewGenericMsg, NewGenericObj },
		{ "InvSelect", NewGenericMsg, NewGenericObj },
		{ "KeypadDone", NewKeypadMsg, NewKeypadObj },
		{ "MediumTransition", NewMediumTransMsg, NewMediumTransObj },
		{ "MotionEnd", NewMotionMsg, NewMotionObj },
		{ "MotionFlagReached", NewMotionMsg, NewMotionObj },
		{ "MotionStart", NewMotionMsg, NewMotionObj },
		{ "MovingTerrainWaypoint", NewMovingTerrainMsg, NewMovingTerrainObj },
		{ "ObjActResult", NewObjActResultMsg, NewObjActResultObj },
		{ "ObjectRoomEnter", NewRoomMsg, NewRoomObj },
		{ "ObjectRoomExit", NewRoomMsg, NewRoomObj },
		{ "ObjRoomTransit", NewGenericMsg, NewGenericObj },
		{ "PatrolPoint", NewPatrolPointMsg, NewPatrolPointObj },
		{ "PhysCollision", NewPhysMsg, NewPhysObj },
		{ "PhysContactCreate", NewPhysMsg, NewPhysObj },
		{ "PhysContactDestroy", NewPhysMsg, NewPhysObj },
		{ "PhysEnter", NewPhysMsg, NewPhysObj },
		{ "PhysExit", NewPhysMsg, NewPhysObj },
		{ "PhysFellAsleep",	NewPhysMsg, NewPhysObj },
		{ "PhysMadeNonPhysical", NewPhysMsg, NewPhysObj },
		{ "PhysMadePhysical", NewPhysMsg, NewPhysObj },
		{ "PhysWokeUp", NewPhysMsg, NewPhysObj },
		{ "PickStateChange", NewPickStateChangeMsg, NewPickStateChangeObj },
		{ "PlayerRoomEnter", NewRoomMsg, NewRoomObj },
		{ "PlayerRoomExit", NewRoomMsg, NewRoomObj },
		{ "PressurePlateActivating", NewPressurePlateMsg, NewPressurePlateObj },
		{ "PressurePlateActive", NewPressurePlateMsg, NewPressurePlateObj },
		{ "PressurePlateDeactivating", NewPressurePlateMsg, NewPressurePlateObj },
		{ "PressurePlateInactive", NewPressurePlateMsg, NewPressurePlateObj },
		{ "QuestChange", NewQuestChangeMsg, NewQuestChangeObj },
		{ "RemotePlayerRoomEnter", NewRoomMsg, NewRoomObj },
		{ "RemotePlayerRoomExit", NewRoomMsg, NewRoomObj },
		{ "SchemaDone", NewSchemaDoneMsg, NewSchemaDoneObj },
		{ "SignalAI", NewSignalAIMsg, NewSignalAIObj },
		{ "Sim", NewSimMsg, NewSimObj },
		{ "Slain", NewSlainMsg, NewSlainObj },
		{ "SoundDone", NewSoundDoneMsg, NewSoundDoneObj },
		{ "StartAttack", NewAttackMsg, NewAttackObj },
		{ "StartWindup", NewAttackMsg, NewAttackObj },
		{ "Timer", NewTimerMsg, NewTimerObj },
		{ "TurnOff", NewGenericMsg, NewGenericObj },
		{ "TurnOn", NewGenericMsg, NewGenericObj },
		{ "TweqComplete", NewTweqCompleteMsg, NewTweqCompleteObj },
		{ "Waypoint", NewWaypointMsg, NewWaypointObj },
		{ "WaypointReached", NewWaypointMsg, NewWaypointObj },
		{ "WorldDeFocus", NewGenericMsg, NewGenericObj },
		{ "WorldDeSelect", NewGenericMsg, NewGenericObj },
		{ "WorldFocus", NewGenericMsg, NewGenericObj },
		{ "WorldSelect", NewGenericMsg, NewGenericObj },
		{ "YorNDone", NewYorNMsg, NewYorNObj },
	};
	const int 
	ScriptMessages::NUM_MSGTYPES = sizeof(_MsgTypes)/sizeof(_MsgTypes[0]);
	const ScriptMessages::_MsgTypeStruct*
	ScriptMessages::_MsgLast = _MsgTypes + NUM_MSGTYPES;

	
	/* To simplify things, these functions do not fail because of the arguments
	 * (or lack thereof). 
	 * The msgName should be a static or interned string.
	 * Also, field names should be the same as declared in <lg/scrmsgs.h>
	 */

	sScrMsg * ScriptMessages::NewGenericMsg(PyObject *)
	{
		return new sScrMsg();
	}

	void ScriptMessages::NewGenericObj(const sScrMsg *, PyObject *)
	{
		// pass
	}

	sScrMsg * ScriptMessages::NewSimMsg(PyObject *msgDict)
	{
		sSimMsg *ret = new sSimMsg();
		PyObject *obStarting = PyDict_GetItemString(msgDict, "fStarting");
		ret->fStarting = (obStarting) ? PyObject_IsTrue(obStarting) : FALSE;
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewSimObj(const sScrMsg *msg, PyObject *msgDict)
	{
		if (static_cast<const sSimMsg*>(msg)->fStarting)
			PyDict_SetItemString(msgDict, "fStarting", Py_True);
		else
			PyDict_SetItemString(msgDict, "fStarting", Py_False);
	}

	sScrMsg * ScriptMessages::NewAIModeChangeMsg(PyObject *msgDict)
	{
		sAIModeChangeMsg *ret = new sAIModeChangeMsg();
		ret->mode = kAIM_Normal;
		ret->previous_mode = kAIM_Normal;
		PyObject *obMode = PyNumber_Int(PyDict_GetItemString(msgDict, "mode"));
		if (obMode)
		{
			ret->mode = static_cast<eAIMode>(PyInt_AsLong(obMode));
			Py_DECREF(obMode);
		}
		obMode = PyNumber_Int(PyDict_GetItemString(msgDict, "previous_mode"));
		if (obMode)
		{
			ret->previous_mode = static_cast<eAIMode>(PyInt_AsLong(obMode));
			Py_DECREF(obMode);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewAIModeChangeObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sAIModeChangeMsg *m = static_cast<const sAIModeChangeMsg*>(msg);
		PyObject *obMode = PyInt_FromLong(m->mode);
		PyDict_SetItemString(msgDict, "mode", obMode);
		Py_DECREF(obMode);
		obMode = PyInt_FromLong(m->previous_mode);
		PyDict_SetItemString(msgDict, "previous_mode", obMode);
		Py_DECREF(obMode);
	}

	sScrMsg * ScriptMessages::NewAlertnessMsg(PyObject *msgDict)
	{
		sAIAlertnessMsg *ret = new sAIAlertnessMsg();
		ret->level = kNoAlert;
		ret->oldLevel = kNoAlert;
		PyObject *obMode = PyNumber_Int(PyDict_GetItemString(msgDict, "level"));
		if (obMode)
		{
			ret->level = static_cast<eAIScriptAlertLevel>(PyInt_AsLong(obMode));
			Py_DECREF(obMode);
		}
		obMode = PyNumber_Int(PyDict_GetItemString(msgDict, "oldLevel"));
		if (obMode)
		{
			ret->oldLevel = static_cast<eAIScriptAlertLevel>(PyInt_AsLong(obMode));
			Py_DECREF(obMode);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewAlertnessObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sAIAlertnessMsg *m = static_cast<const sAIAlertnessMsg*>(msg);
		PyObject *obMode = PyInt_FromLong(m->level);
		PyDict_SetItemString(msgDict, "level", obMode);
		Py_DECREF(obMode);
		obMode = PyInt_FromLong(m->oldLevel);
		PyDict_SetItemString(msgDict, "oldLevel", obMode);
		Py_DECREF(obMode);
	}

	sScrMsg * ScriptMessages::NewHighAlertMsg(PyObject *msgDict)
	{
		sAIHighAlertMsg *ret = new sAIHighAlertMsg();
		ret->level = kNoAlert;
		ret->oldLevel = kNoAlert;
		PyObject *obMode = PyNumber_Int(PyDict_GetItemString(msgDict, "level"));
		if (obMode)
		{
			ret->level = static_cast<eAIScriptAlertLevel>(PyInt_AsLong(obMode));
			Py_DECREF(obMode);
		}
		obMode = PyNumber_Int(PyDict_GetItemString(msgDict, "oldLevel"));
		if (obMode)
		{
			ret->oldLevel = static_cast<eAIScriptAlertLevel>(PyInt_AsLong(obMode));
			Py_DECREF(obMode);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewHighAlertObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sAIHighAlertMsg *m = static_cast<const sAIHighAlertMsg*>(msg);
		PyObject *obMode = PyInt_FromLong(m->level);
		PyDict_SetItemString(msgDict, "level", obMode);
		Py_DECREF(obMode);
		obMode = PyInt_FromLong(m->oldLevel);
		PyDict_SetItemString(msgDict, "oldLevel", obMode);
		Py_DECREF(obMode);
	}

	sScrMsg * ScriptMessages::NewObjActResultMsg(PyObject *msgDict)
	{
		sAIObjActResultMsg *ret = new sAIObjActResultMsg();
		ret->target = 0;
		ret->action = kAINoAction;
		ret->result = kActionDone;
		PyObject *obTarget = PyNumber_Int(PyDict_GetItemString(msgDict, "target"));
		if (obTarget)
		{
			ret->target = PyInt_AsLong(obTarget);
			Py_DECREF(obTarget);
		}
		PyObject *obAction = PyNumber_Int(PyDict_GetItemString(msgDict, "action"));
		if (obAction)
		{
			ret->action = static_cast<eAIAction>(PyInt_AsLong(obAction));
			Py_DECREF(obAction);
		}
		PyObject *obResult = PyNumber_Int(PyDict_GetItemString(msgDict, "result"));
		if (obResult)
		{
			ret->result = static_cast<eAIActionResult>(PyInt_AsLong(obResult));
			Py_DECREF(obResult);
		}
		PyObject *obData = PyDict_GetItemString(msgDict, "result_data");
		if (obData)
		{
			MultiParmFromObject(ret->result_data, obData);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewObjActResultObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sAIObjActResultMsg *m = static_cast<const sAIObjActResultMsg*>(msg);
		PyObject *obTarget = ObjectObject::New(m->target);
		PyDict_SetItemString(msgDict, "target", obTarget);
		Py_DECREF(obTarget);
		PyObject *obAction = PyInt_FromLong(m->action);
		PyDict_SetItemString(msgDict, "action", obAction);
		Py_DECREF(obAction);
		PyObject *obResult = PyInt_FromLong(m->result);
		PyDict_SetItemString(msgDict, "result", obResult);
		Py_DECREF(obResult);
		PyObject *obData = MultiParmToObject(m->result_data);
		if (obData && obData != Py_None)
		{
			PyDict_SetItemString(msgDict, "result_data", obData);
		}
		Py_XDECREF(obData);
	}

	sScrMsg * ScriptMessages::NewPatrolPointMsg(PyObject *msgDict)
	{
		sAIPatrolPointMsg *ret = new sAIPatrolPointMsg();
		ret->patrolObj = 0;
		PyObject *obTarget = PyNumber_Int(PyDict_GetItemString(msgDict, "patrolObj"));
		if (obTarget)
		{
			ret->patrolObj = PyInt_AsLong(obTarget);
			Py_DECREF(obTarget);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewPatrolPointObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sAIPatrolPointMsg *m = static_cast<const sAIPatrolPointMsg*>(msg);
		PyObject *obTarget = ObjectObject::New(m->patrolObj);
		PyDict_SetItemString(msgDict, "patrlObj", obTarget);
		Py_DECREF(obTarget);
	}

	sScrMsg * ScriptMessages::NewSignalAIMsg(PyObject *msgDict)
	{
		sAISignalMsg *ret = new sAISignalMsg();
		PyObject *obSignal = PyDict_GetItemString(msgDict, "signal");
		if (obSignal)
		{
			ret->signal = GetInternString(obSignal);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewSignalAIObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sAISignalMsg *m = static_cast<const sAISignalMsg*>(msg);
		PyObject *obSignal = PyString_FromString(m->signal);
		PyDict_SetItemString(msgDict, "signal", obSignal);
		Py_DECREF(obSignal);
	}

	sScrMsg * ScriptMessages::NewAttackMsg(PyObject *msgDict)
	{
		sAttackMsg *ret = new sAttackMsg();
		ret->weapon = 0;
		PyObject *obWeapon = PyNumber_Int(PyDict_GetItemString(msgDict, "weapon"));
		if (obWeapon)
		{
			ret->weapon = PyInt_AsLong(obWeapon);
			Py_DECREF(obWeapon);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewAttackObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sAttackMsg *m = static_cast<const sAttackMsg*>(msg);
		PyObject *obWeapon = ObjectObject::New(m->weapon);
		PyDict_SetItemString(msgDict, "weapon", obWeapon);
		Py_DECREF(obWeapon);
	}

	sScrMsg * ScriptMessages::NewCombineMsg(PyObject *msgDict)
	{
		sCombineScrMsg *ret = new sCombineScrMsg();
		ret->combiner = 0;
		PyObject *obCombiner = PyNumber_Int(PyDict_GetItemString(msgDict, "combiner"));
		if (obCombiner)
		{
			ret->combiner = PyInt_AsLong(obCombiner);
			Py_DECREF(obCombiner);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewCombineObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sCombineScrMsg *m = static_cast<const sCombineScrMsg*>(msg);
		PyObject *obCombiner = ObjectObject::New(m->combiner);
		PyDict_SetItemString(msgDict, "combiner", obCombiner);
		Py_DECREF(obCombiner);
	}

	sScrMsg * ScriptMessages::NewContainedMsg(PyObject *msgDict)
	{
		sContainedScrMsg *ret = new sContainedScrMsg();
		ret->event = 0;
		ret->container = 0;
		PyObject *obEvent = PyNumber_Int(PyDict_GetItemString(msgDict, "event"));
		if (obEvent)
		{
			ret->event = PyInt_AsLong(obEvent);
			Py_DECREF(obEvent);
		}
		PyObject *obCont = PyNumber_Int(PyDict_GetItemString(msgDict, "container"));
		if (obCont)
		{
			ret->container = PyInt_AsLong(obCont);
			Py_DECREF(obCont);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewContainedObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sContainedScrMsg *m = static_cast<const sContainedScrMsg*>(msg);
		PyObject *obCont = ObjectObject::New(m->container);
		PyDict_SetItemString(msgDict, "container", obCont);
		Py_DECREF(obCont);
		PyObject *obEvent = PyInt_FromLong(m->event);
		PyDict_SetItemString(msgDict, "event", obEvent);
		Py_DECREF(obEvent);
	}

	sScrMsg * ScriptMessages::NewContainerMsg(PyObject *msgDict)
	{
		sContainerScrMsg *ret = new sContainerScrMsg();
		ret->event = 0;
		ret->containee = 0;
		PyObject *obEvent = PyNumber_Int(PyDict_GetItemString(msgDict, "event"));
		if (obEvent)
		{
			ret->event = PyInt_AsLong(obEvent);
			Py_DECREF(obEvent);
		}
		PyObject *obCont = PyNumber_Int(PyDict_GetItemString(msgDict, "containee"));
		if (obCont)
		{
			ret->containee = PyInt_AsLong(obCont);
			Py_DECREF(obCont);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewContainerObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sContainerScrMsg *m = static_cast<const sContainerScrMsg*>(msg);
		PyObject *obCont = ObjectObject::New(m->containee);
		PyDict_SetItemString(msgDict, "containee", obCont);
		Py_DECREF(obCont);
		PyObject *obEvent = PyInt_FromLong(m->event);
		PyDict_SetItemString(msgDict, "event", obEvent);
		Py_DECREF(obEvent);
	}

	sScrMsg * ScriptMessages::NewDamageMsg(PyObject *msgDict)
	{
		sDamageScrMsg *ret = new sDamageScrMsg();
		ret->kind = 0;
		ret->damage = 0;
		ret->culprit = 0;
		PyObject *obDamage = PyNumber_Int(PyDict_GetItemString(msgDict, "damage"));
		if (obDamage)
		{
			ret->damage = PyInt_AsLong(obDamage);
			Py_DECREF(obDamage);
		}
		PyObject *obCulprit = PyNumber_Int(PyDict_GetItemString(msgDict, "culprit"));
		if (obCulprit)
		{
			ret->culprit = PyInt_AsLong(obCulprit);
			Py_DECREF(obCulprit);
		}
		PyObject *obKind = PyNumber_Int(PyDict_GetItemString(msgDict, "kind"));
		if (obKind)
		{
			ret->kind = PyInt_AsLong(obKind);
			Py_DECREF(obKind);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewDamageObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sDamageScrMsg *m = static_cast<const sDamageScrMsg*>(msg);
		PyObject *obDamage = PyInt_FromLong(m->damage);
		PyDict_SetItemString(msgDict, "damage", obDamage);
		Py_DECREF(obDamage);
		PyObject *obCulprit = ObjectObject::New(m->culprit);
		PyDict_SetItemString(msgDict, "culprit", obCulprit);
		Py_DECREF(obCulprit);
		PyObject *obKind = PyInt_FromLong(m->kind);
		PyDict_SetItemString(msgDict, "kind", obKind);
		Py_DECREF(obKind);
	}

	sScrMsg * ScriptMessages::NewDifficultyMsg(PyObject *msgDict)
	{
		sDiffScrMsg *ret = new sDiffScrMsg();
		ret->difficulty = 0;
		PyObject *obDiff = PyNumber_Int(PyDict_GetItemString(msgDict, "difficulty"));
		if (obDiff)
		{
			ret->difficulty = PyInt_AsLong(obDiff);
			Py_DECREF(obDiff);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewDifficultyObj(const sScrMsg *msg, PyObject *msgDict)
	{
		PyObject *obDiff = PyInt_FromLong(static_cast<const sDiffScrMsg*>(msg)->difficulty);
		PyDict_SetItemString(msgDict, "difficulty", obDiff);
		Py_DECREF(obDiff);
	}

	sScrMsg * ScriptMessages::NewDoorMsg(PyObject *msgDict)
	{
		sDoorMsg *ret = new sDoorMsg();
		ret->ActionType = sDoorMsg::kOpen;
		ret->PrevActionType = sDoorMsg::kOpen;
		PyObject *obAction = PyNumber_Int(PyDict_GetItemString(msgDict, "ActionType"));
		if (obAction)
		{
			ret->ActionType = static_cast< sDoorMsg::eDoorAction>(PyInt_AsLong(obAction));
			Py_DECREF(obAction);
		}
		obAction = PyNumber_Int(PyDict_GetItemString(msgDict, "PrevActionType"));
		if (obAction)
		{
			ret->PrevActionType = static_cast< sDoorMsg::eDoorAction>(PyInt_AsLong(obAction));
			Py_DECREF(obAction);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewDoorObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sDoorMsg *m = static_cast<const sDoorMsg*>(msg);
		PyObject *obAction = PyInt_FromLong(m->ActionType);
		PyDict_SetItemString(msgDict, "ActionType", obAction);
		Py_DECREF(obAction);
		obAction = PyInt_FromLong(m->PrevActionType);
		PyDict_SetItemString(msgDict, "PrevActionType", obAction);
		Py_DECREF(obAction);
	}

	sScrMsg * ScriptMessages::NewFrobMsg(PyObject *msgDict)
	{
		sFrobMsg *ret = new sFrobMsg();
		ret->SrcObjId = 0;
		ret->DstObjId = 0;
		ret->Frobber = 0;
		ret->SrcLoc = kFrobLocWorld;
		ret->DstLoc = kFrobLocWorld;
		ret->Sec = 0;
		ret->Abort = 0;
		PyObject *obObjId = PyNumber_Int(PyDict_GetItemString(msgDict, "SrcObjId"));
		if (obObjId)
		{
			ret->SrcObjId = PyInt_AsLong(obObjId);
			Py_DECREF(obObjId);
		}
		obObjId = PyNumber_Int(PyDict_GetItemString(msgDict, "DstObjId"));
		if (obObjId)
		{
			ret->DstObjId = PyInt_AsLong(obObjId);
			Py_DECREF(obObjId);
		}
		obObjId = PyNumber_Int(PyDict_GetItemString(msgDict, "Frobber"));
		if (obObjId)
		{
			ret->Frobber = PyInt_AsLong(obObjId);
			Py_DECREF(obObjId);
		}
		PyObject *obLoc = PyNumber_Int(PyDict_GetItemString(msgDict, "SrcLoc"));
		if (obLoc)
		{
			ret->SrcLoc = static_cast<eFrobLoc>(PyInt_AsLong(obLoc));
			Py_DECREF(obLoc);
		}
		obLoc = PyNumber_Int(PyDict_GetItemString(msgDict, "DstLoc"));
		if (obLoc)
		{
			ret->DstLoc = static_cast<eFrobLoc>(PyInt_AsLong(obLoc));
			Py_DECREF(obLoc);
		}
		PyObject *obSec = PyNumber_Float(PyDict_GetItemString(msgDict, "Sec"));
		if (obSec)
		{
			ret->Sec = PyFloat_AsDouble(obSec);
			Py_DECREF(obSec);
		}
		PyObject *obFlag = PyNumber_Int(PyDict_GetItemString(msgDict, "Abort"));
		if (obFlag)
		{
			ret->Abort = PyInt_AsLong(obFlag);
			Py_DECREF(obFlag);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewFrobObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sFrobMsg *m = static_cast<const sFrobMsg*>(msg);
		PyObject *obObj = ObjectObject::New(m->Frobber);
		PyDict_SetItemString(msgDict, "Frobber", obObj);
		Py_DECREF(obObj);
		obObj = ObjectObject::New(m->SrcObjId);
		PyDict_SetItemString(msgDict, "SrcObjId", obObj);
		Py_DECREF(obObj);
		obObj = ObjectObject::New(m->DstObjId);
		PyDict_SetItemString(msgDict, "DstObjId", obObj);
		Py_DECREF(obObj);
		PyObject *obLoc = PyInt_FromLong(m->SrcLoc);
		PyDict_SetItemString(msgDict, "SrcLoc", obLoc);
		Py_DECREF(obLoc);
		obLoc = PyInt_FromLong(m->DstLoc);
		PyDict_SetItemString(msgDict, "DstLoc", obLoc);
		Py_DECREF(obLoc);
		PyObject *obFlag = (m->Abort) ? Py_True : Py_False;
		PyDict_SetItemString(msgDict, "Abort", obFlag);
	}

	sScrMsg * ScriptMessages::NewMotionMsg(PyObject *msgDict)
	{
		sBodyMsg *ret = new sBodyMsg();
		ret->ActionType = sBodyMsg::kMotionStart;
		ret->FlagValue = 0;
		PyObject *obAction = PyNumber_Int(PyDict_GetItemString(msgDict, "ActionType"));
		if (obAction)
		{
			ret->ActionType = static_cast<sBodyMsg::eBodyAction>(PyInt_AsLong(obAction));
			Py_DECREF(obAction);
		}
		PyObject *obFlag = PyNumber_Int(PyDict_GetItemString(msgDict, "FlagValue"));
		if (obFlag)
		{
			ret->FlagValue = PyInt_AsLong(obFlag);
			Py_DECREF(obFlag);
		}
		PyObject *obName = PyDict_GetItemString(msgDict, "MotionName");
		if (obName)
		{
			ret->MotionName = GetInternString(obName);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewMotionObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sBodyMsg *m = static_cast<const sBodyMsg*>(msg);
		PyObject *obName = PyString_FromString(m->MotionName);
		PyDict_SetItemString(msgDict, "MotionName", obName);
		Py_DECREF(obName);
		PyObject *obAction = PyInt_FromLong(m->ActionType);
		PyDict_SetItemString(msgDict, "ActionType", obAction);
		Py_DECREF(obAction);
		PyObject *obFlag = PyInt_FromLong(m->FlagValue);
		PyDict_SetItemString(msgDict, "FlagValue", obFlag);
		Py_DECREF(obFlag);
	}

	sScrMsg * ScriptMessages::NewPickStateChangeMsg(PyObject *msgDict)
	{
		sPickStateScrMsg *ret = new sPickStateScrMsg();
		ret->PrevState = 0;
		ret->NewState = 0;
		PyObject *obState = PyNumber_Int(PyDict_GetItemString(msgDict, "NewState"));
		if (obState)
		{
			ret->NewState = PyInt_AsLong(obState);
			Py_DECREF(obState);
		}
		obState = PyNumber_Int(PyDict_GetItemString(msgDict, "PrevState"));
		if (obState)
		{
			ret->PrevState = PyInt_AsLong(obState);
			Py_DECREF(obState);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewPickStateChangeObj(const sScrMsg *msg, PyObject *msgDict)
	{
		PyObject *obState = PyInt_FromLong(static_cast<const sPickStateScrMsg*>(msg)->NewState);
		PyDict_SetItemString(msgDict, "NewState", obState);
		Py_DECREF(obState);
		obState = PyInt_FromLong(static_cast<const sPickStateScrMsg*>(msg)->PrevState);
		PyDict_SetItemString(msgDict, "PrevState", obState);
		Py_DECREF(obState);
	}

	sScrMsg * ScriptMessages::NewPressurePlateMsg(PyObject *)
	{
		sPressurePlateMsg *ret = new sPressurePlateMsg();
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewPressurePlateObj(const sScrMsg *, PyObject *)
	{
		// Nothing to do.
	}

	sScrMsg * ScriptMessages::NewPhysMsg(PyObject *msgDict)
	{
		sPhysMsg *ret = new sPhysMsg();
		memset(&ret->Submod, 0, sizeof(sPhysMsg) - sizeof(sScrMsg));

		PyObject *obType = PyNumber_Int(PyDict_GetItemString(msgDict, "collType"));
		if (obType)
		{
			ret->collType = PyInt_AsLong(obType);
			Py_DECREF(obType);
		}
		obType = PyNumber_Int(PyDict_GetItemString(msgDict, "contactType"));
		if (obType)
		{
			ret->contactType = PyInt_AsLong(obType);
			Py_DECREF(obType);
		}
		obType = PyNumber_Int(PyDict_GetItemString(msgDict, "Submod"));
		if (obType)
		{
			ret->Submod = PyInt_AsLong(obType);
			Py_DECREF(obType);
		}
		obType = PyNumber_Int(PyDict_GetItemString(msgDict, "collSubmod"));
		if (obType)
		{
			ret->collSubmod = PyInt_AsLong(obType);
			Py_DECREF(obType);
		}
		obType = PyNumber_Int(PyDict_GetItemString(msgDict, "contactSubmod"));
		if (obType)
		{
			ret->contactSubmod = PyInt_AsLong(obType);
			Py_DECREF(obType);
		}
		obType = PyNumber_Int(PyDict_GetItemString(msgDict, "transSubmod"));
		if (obType)
		{
			ret->transSubmod = PyInt_AsLong(obType);
			Py_DECREF(obType);
		}
		PyObject *obObj = PyNumber_Int(PyDict_GetItemString(msgDict, "collObj"));
		if (obObj)
		{
			ret->collObj = PyInt_AsLong(obObj);
			Py_DECREF(obObj);
		}
		obObj = PyNumber_Int(PyDict_GetItemString(msgDict, "contactObj"));
		if (obObj)
		{
			ret->contactObj = PyInt_AsLong(obObj);
			Py_DECREF(obObj);
		}
		obObj = PyNumber_Int(PyDict_GetItemString(msgDict, "transObj"));
		if (obObj)
		{
			ret->transObj = PyInt_AsLong(obObj);
			Py_DECREF(obObj);
		}
		PyObject *obMmnt = PyNumber_Float(PyDict_GetItemString(msgDict, "collMomentum"));
		if (obMmnt)
		{
			ret->collMomentum = PyFloat_AsDouble(obMmnt);
			Py_DECREF(obMmnt);
		}
		PyObject *obPt = PyDict_GetItemString(msgDict, "collPt");
		if (obPt && PySequence_Check(obPt) && PySequence_Size(obPt) == 3)
		{
			obPt = PySequence_Tuple(obPt);
			PyArg_ParseTuple(obPt, "fff", &ret->collPt.x, &ret->collPt.y, &ret->collPt.z);
			Py_DECREF(obPt);
			PyErr_Clear();
		}
		obPt = PyDict_GetItemString(msgDict, "collNormal");
		if (obPt && PySequence_Check(obPt) && PySequence_Size(obPt) == 3)
		{
			obPt = PySequence_Tuple(obPt);
			PyArg_ParseTuple(obPt, "fff", &ret->collNormal.x, &ret->collNormal.y, &ret->collNormal.z);
			Py_DECREF(obPt);
			PyErr_Clear();
		}

		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewPhysObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sPhysMsg *m = static_cast<const sPhysMsg*>(msg);
		PyObject *obType = PyInt_FromLong(m->collType);
		PyDict_SetItemString(msgDict, "collType", obType);
		Py_DECREF(obType);
		obType =  PyInt_FromLong(m->contactType);
		PyDict_SetItemString(msgDict, "contactType", obType);
		Py_DECREF(obType);
		obType =  PyInt_FromLong(m->collSubmod);
		PyDict_SetItemString(msgDict, "collSubmod", obType);
		Py_DECREF(obType);
		obType =  PyInt_FromLong(m->contactSubmod);
		PyDict_SetItemString(msgDict, "contactSubmod", obType);
		Py_DECREF(obType);
		obType =  PyInt_FromLong(m->transSubmod);
		PyDict_SetItemString(msgDict, "transSubmod", obType);
		Py_DECREF(obType);
		obType =  PyInt_FromLong(m->Submod);
		PyDict_SetItemString(msgDict, "Submod", obType);
		Py_DECREF(obType);
		PyObject *obObj = ObjectObject::New(m->collObj);
		PyDict_SetItemString(msgDict, "collObj", obObj);
		Py_DECREF(obObj);
		obObj = ObjectObject::New(m->contactObj);
		PyDict_SetItemString(msgDict, "contactObj", obObj);
		Py_DECREF(obObj);
		obObj = ObjectObject::New(m->transObj);
		PyDict_SetItemString(msgDict, "transObj", obObj);
		Py_DECREF(obObj);
		PyObject *obMmnt = PyFloat_FromDouble(m->collMomentum);
		PyDict_SetItemString(msgDict, "collMomentum", obMmnt);
		Py_DECREF(obMmnt);
		PyObject *obVect = Py_BuildValue("[fff]", m->collPt.x, m->collPt.y, m->collPt.z);
		PyDict_SetItemString(msgDict, "collPt", obVect);
		Py_DECREF(obVect);
		obVect = Py_BuildValue("[fff]", m->collNormal.x, m->collNormal.y, m->collNormal.z);
		PyDict_SetItemString(msgDict, "collNormal", obVect);
		Py_DECREF(obVect);
	}

	sScrMsg * ScriptMessages::NewRoomMsg(PyObject *msgDict)
	{
		sRoomMsg *ret = new sRoomMsg();
		ret->FromObjId = 0;
		ret->ToObjId = 0;
		ret->MoveObjId = 0;
		ret->ObjType = sRoomMsg::kNull;
		ret->TransitionType = sRoomMsg::kEnter;
		PyObject *obObj = PyNumber_Int(PyDict_GetItemString(msgDict, "FromObjId"));
		if (obObj)
		{
			ret->FromObjId = PyInt_AsLong(obObj);
			Py_DECREF(obObj);
		}
		obObj = PyNumber_Int(PyDict_GetItemString(msgDict, "ToObjId"));
		if (obObj)
		{
			ret->ToObjId = PyInt_AsLong(obObj);
			Py_DECREF(obObj);
		}
		obObj = PyNumber_Int(PyDict_GetItemString(msgDict, "MoveObjId"));
		if (obObj)
		{
			ret->MoveObjId = PyInt_AsLong(obObj);
			Py_DECREF(obObj);
		}
		PyObject *obType = PyNumber_Int(PyDict_GetItemString(msgDict, "ObjType"));
		if (obType)
		{
			ret->ObjType = static_cast<sRoomMsg::eObjType>(PyInt_AsLong(obType));
			Py_DECREF(obType);
		}
		obType = PyNumber_Int(PyDict_GetItemString(msgDict, "TransitionType"));
		if (obType)
		{
			ret->TransitionType = static_cast<sRoomMsg::eTransType>(PyInt_AsLong(obType));
			Py_DECREF(obType);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewRoomObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sRoomMsg *m = static_cast<const sRoomMsg*>(msg);
		PyObject *obObj = ObjectObject::New(m->FromObjId);
		PyDict_SetItemString(msgDict, "FromObjId", obObj);
		Py_DECREF(obObj);
		obObj = ObjectObject::New(m->ToObjId);
		PyDict_SetItemString(msgDict, "ToObjId", obObj);
		Py_DECREF(obObj);
		obObj = ObjectObject::New(m->MoveObjId);
		PyDict_SetItemString(msgDict, "MoveObjId", obObj);
		Py_DECREF(obObj);
		PyObject *obType = PyInt_FromLong(m->ObjType);
		PyDict_SetItemString(msgDict, "ObjType", obType);
		Py_DECREF(obType);
		obType = PyInt_FromLong(m->TransitionType);
		PyDict_SetItemString(msgDict, "TransitionType", obType);
		Py_DECREF(obType);

	}

	sScrMsg * ScriptMessages::NewSlainMsg(PyObject *msgDict)
	{
		sSlayMsg *ret = new sSlayMsg();
		ret->culprit = 0;
		ret->kind = 0;
		PyObject *obObj = PyNumber_Int(PyDict_GetItemString(msgDict, "culprit"));
		if (obObj)
		{
			ret->culprit = PyInt_AsLong(obObj);
			Py_DECREF(obObj);
		}
		PyObject *obType = PyNumber_Int(PyDict_GetItemString(msgDict, "kind"));
		if (obType)
		{
			ret->kind = PyInt_AsLong(obType);
			Py_DECREF(obType);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewSlainObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sSlayMsg *m = static_cast<const sSlayMsg*>(msg);
		PyObject *obObj = ObjectObject::New(m->culprit);
		PyDict_SetItemString(msgDict, "culprit", obObj);
		Py_DECREF(obObj);
		PyObject *obKind = PyInt_FromLong(m->kind);
		PyDict_SetItemString(msgDict, "kind", obKind);
		Py_DECREF(obKind);
	}

	sScrMsg * ScriptMessages::NewSchemaDoneMsg(PyObject *msgDict)
	{
		sSchemaDoneMsg *ret = new sSchemaDoneMsg();
		ret->targetObject = 0;
		PyObject *obObj = PyNumber_Int(PyDict_GetItemString(msgDict, "targetObject"));
		if (obObj)
		{
			ret->targetObject = PyInt_AsLong(obObj);
			Py_DECREF(obObj);
		}
		PyObject *obVect = PyDict_GetItemString(msgDict, "coordinates");
		if (obVect && PySequence_Check(obVect) && PySequence_Size(obVect) == 3)
		{
			obVect = PySequence_Tuple(obVect);
			PyArg_ParseTuple(obVect, "fff", &ret->coordinates.x, &ret->coordinates.y, &ret->coordinates.z);
			Py_DECREF(obVect);
			PyErr_Clear();
		}
		PyObject *obName = PyDict_GetItemString(msgDict, "name");
		if (obName)
		{
			ret->name = GetInternString(obName);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewSchemaDoneObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sSchemaDoneMsg *m = static_cast<const sSchemaDoneMsg*>(msg);
		PyObject *obName = PyString_FromString(m->name);
		PyDict_SetItemString(msgDict, "name", obName);
		Py_DECREF(obName);
		PyObject *obVect = Py_BuildValue("[fff]", m->coordinates.x, m->coordinates.y, m->coordinates.z);
		PyDict_SetItemString(msgDict, "coordinates", obVect);
		Py_DECREF(obVect);
		PyObject *obTarget = ObjectObject::New(m->targetObject);
		PyDict_SetItemString(msgDict, "targetObject", obTarget);
		Py_DECREF(obTarget);
	}

	sScrMsg * ScriptMessages::NewSoundDoneMsg(PyObject *msgDict)
	{
		sSoundDoneMsg *ret = new sSoundDoneMsg();
		ret->targetObject = 0;
		PyObject *obObj = PyNumber_Int(PyDict_GetItemString(msgDict, "targetObject"));
		if (obObj)
		{
			ret->targetObject = PyInt_AsLong(obObj);
			Py_DECREF(obObj);
		}
		PyObject *obVect = PyDict_GetItemString(msgDict, "coordinates");
		if (obVect && PySequence_Check(obVect) && PySequence_Size(obVect) == 3)
		{
			obVect = PySequence_Tuple(obVect);
			PyArg_ParseTuple(obVect, "fff", &ret->coordinates.x, &ret->coordinates.y, &ret->coordinates.z);
			Py_DECREF(obVect);
			PyErr_Clear();
		}
		PyObject *obName = PyDict_GetItemString(msgDict, "name");
		if (obName)
		{
			ret->name = GetInternString(obName);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewSoundDoneObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sSoundDoneMsg *m = static_cast<const sSoundDoneMsg*>(msg);
		PyObject *obName = PyString_FromString(m->name);
		PyDict_SetItemString(msgDict, "name", obName);
		Py_DECREF(obName);
		PyObject *obVect = Py_BuildValue("[fff]", m->coordinates.x, m->coordinates.y, m->coordinates.z);
		PyDict_SetItemString(msgDict, "coordinates", obVect);
		Py_DECREF(obVect);
		PyObject *obTarget = ObjectObject::New(m->targetObject);
		PyDict_SetItemString(msgDict, "targetObject", obTarget);
		Py_DECREF(obTarget);
	}

	sScrMsg * ScriptMessages::NewStimMsg(PyObject *msgDict)
	{
		sStimMsg *ret = new sStimMsg();
		ret->intensity = 0;
		ret->sensor = 0;
		ret->source = 0;
		ret->stimulus = 0;
		PyObject *obMag = PyNumber_Float(PyDict_GetItemString(msgDict, "intensity"));
		if (obMag)
		{
			ret->intensity = PyFloat_AsDouble(obMag);
			Py_DECREF(obMag);
		}
		PyObject *obObj = PyNumber_Int(PyDict_GetItemString(msgDict, "stimulus"));
		if (obObj)
		{
			ret->stimulus = PyInt_AsLong(obObj);
			Py_DECREF(obObj);
		}
		obObj = PyNumber_Int(PyDict_GetItemString(msgDict, "sensor"));
		if (obObj)
		{
			ret->sensor = PyInt_AsLong(obObj);
			Py_DECREF(obObj);
		}
		obObj = PyNumber_Int(PyDict_GetItemString(msgDict, "source"));
		if (obObj)
		{
			ret->source = PyInt_AsLong(obObj);
			Py_DECREF(obObj);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewStimObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sStimMsg *m = static_cast<const sStimMsg*>(msg);
		PyObject *obMag = PyFloat_FromDouble(m->intensity);
		PyDict_SetItemString(msgDict, "intensity", obMag);
		Py_DECREF(obMag);
		PyObject *obObj = ObjectObject::New(m->stimulus);
		PyDict_SetItemString(msgDict, "stimulus", obObj);
		Py_DECREF(obObj);
		obObj = LinkObject::New(m->sensor);
		PyDict_SetItemString(msgDict, "sensor", obObj);
		Py_DECREF(obObj);
		obObj = LinkObject::New(m->source);
		PyDict_SetItemString(msgDict, "source", obObj);
		Py_DECREF(obObj);
	}

	sScrMsg * ScriptMessages::NewTimerMsg(PyObject *msgDict)
	{
		sScrTimerMsg *ret = new sScrTimerMsg();
		PyObject *obName = PyDict_GetItemString(msgDict, "name");
		if (obName)
		{
			ret->name = GetInternString(obName);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewTimerObj(const sScrMsg *msg, PyObject *msgDict)
	{
		PyObject *obName = PyString_FromString(static_cast<const sScrTimerMsg*>(msg)->name);
		PyDict_SetItemString(msgDict, "name", obName);
		Py_DECREF(obName);
	}

	sScrMsg * ScriptMessages::NewTweqCompleteMsg(PyObject *msgDict)
	{
		sTweqMsg *ret = new sTweqMsg();
		ret->Type = kTweqTypeScale;
		ret->Op = kTweqOpKillAll;
		ret->Dir = kTweqDirForward;
		PyObject *obType = PyNumber_Int(PyDict_GetItemString(msgDict, "Type"));
		if (obType)
		{
			ret->Type = static_cast<eTweqType>(PyInt_AsLong(obType));
			Py_DECREF(obType);
		}
		obType = PyNumber_Int(PyDict_GetItemString(msgDict, "Op"));
		if (obType)
		{
			ret->Op = static_cast<eTweqOperation>(PyInt_AsLong(obType));
			Py_DECREF(obType);
		}
		obType = PyNumber_Int(PyDict_GetItemString(msgDict, "Dir"));
		if (obType)
		{
			ret->Op = static_cast<eTweqOperation>(PyInt_AsLong(obType));
			Py_DECREF(obType);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewTweqCompleteObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sTweqMsg *m = static_cast<const sTweqMsg*>(msg);
		PyObject *obType = PyInt_FromLong(m->Type);
		PyDict_SetItemString(msgDict, "Type", obType);
		Py_DECREF(obType);
		obType = PyInt_FromLong(m->Op);
		PyDict_SetItemString(msgDict, "Op", obType);
		Py_DECREF(obType);
		obType = PyInt_FromLong(m->Dir);
		PyDict_SetItemString(msgDict, "Dir", obType);
		Py_DECREF(obType);
	}

	sScrMsg * ScriptMessages::NewWaypointMsg(PyObject *msgDict)
	{
		sWaypointMsg *ret = new sWaypointMsg();
		ret->moving_terrain = 0;
		PyObject *obObj = PyNumber_Int(PyDict_GetItemString(msgDict, "moving_terrain"));
		if (obObj)
		{
			ret->moving_terrain = PyInt_AsLong(obObj);
			Py_DECREF(obObj);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewWaypointObj(const sScrMsg *msg, PyObject *msgDict)
	{
		PyObject *obObj = ObjectObject::New(static_cast<const sWaypointMsg*>(msg)->moving_terrain);
		PyDict_SetItemString(msgDict, "moving_terrain", obObj);
		Py_DECREF(obObj);
	}

	sScrMsg * ScriptMessages::NewMovingTerrainMsg(PyObject *msgDict)
	{
		sMovingTerrainMsg *ret = new sMovingTerrainMsg();
		ret->waypoint = 0;
		PyObject *obObj = PyNumber_Int(PyDict_GetItemString(msgDict, "waypoint"));
		if (obObj)
		{
			ret->waypoint = PyInt_AsLong(obObj);
			Py_DECREF(obObj);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewMovingTerrainObj(const sScrMsg *msg, PyObject *msgDict)
	{
		PyObject *obObj = ObjectObject::New(static_cast<const sMovingTerrainMsg*>(msg)->waypoint);
		PyDict_SetItemString(msgDict, "waypoint", obObj);
		Py_DECREF(obObj);
	}

	sScrMsg * ScriptMessages::NewQuestChangeMsg(PyObject *msgDict)
	{
		sQuestMsg *ret = new sQuestMsg();
		ret->m_oldValue = 0;
		ret->m_newValue = 0;
		ret->m_pName = NULL;
		PyObject *obVal = PyNumber_Int(PyDict_GetItemString(msgDict, "m_oldValue"));
		if (obVal)
		{
			ret->m_oldValue = PyInt_AsLong(obVal);
			Py_DECREF(obVal);
		}
		obVal = PyNumber_Int(PyDict_GetItemString(msgDict, "m_newValue"));
		if (obVal)
		{
			ret->m_newValue = PyInt_AsLong(obVal);
			Py_DECREF(obVal);
		}
		PyObject *obName = PyDict_GetItemString(msgDict, "m_pName");
		if (obName)
		{
			ret->m_pName = GetInternString(obName);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewQuestChangeObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sQuestMsg *m = static_cast<const sQuestMsg*>(msg);
		PyObject *obName = PyString_FromString(m->m_pName);
		PyDict_SetItemString(msgDict, "m_pName", obName);
		Py_DECREF(obName);
		PyObject *obVal = PyInt_FromLong(m->m_newValue);
		PyDict_SetItemString(msgDict, "m_newValue", obVal);
		Py_DECREF(obVal);
		obVal = PyInt_FromLong(m->m_oldValue);
		PyDict_SetItemString(msgDict, "m_oldValue", obVal);
		Py_DECREF(obVal);
	}

	sScrMsg * ScriptMessages::NewMediumTransMsg(PyObject *msgDict)
	{
		sMediumTransMsg *ret = new sMediumTransMsg();
		PyObject *obType = PyNumber_Int(PyDict_GetItemString(msgDict, "nFromType"));
		if (obType)
		{
			ret->nFromType = PyInt_AsLong(obType);
			Py_DECREF(obType);
		}
		obType = PyNumber_Int(PyDict_GetItemString(msgDict, "nToType"));
		if (obType)
		{
			ret->nToType = PyInt_AsLong(obType);
			Py_DECREF(obType);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewMediumTransObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sMediumTransMsg *m = static_cast<const sMediumTransMsg*>(msg);
		PyObject *obType = PyInt_FromLong(m->nFromType);
		PyDict_SetItemString(msgDict, "nFromType", obType);
		Py_DECREF(obType);
		obType = PyInt_FromLong(m->nToType);
		PyDict_SetItemString(msgDict, "nToType", obType);
		Py_DECREF(obType);
	}

	sScrMsg * ScriptMessages::NewYorNMsg(PyObject *msgDict)
	{
		sYorNMsg *ret = new sYorNMsg();
		PyObject *obYorN = PyDict_GetItemString(msgDict, "YorN");
		ret->YorN = (obYorN) ? PyObject_IsTrue(obYorN) : FALSE;
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewYorNObj(const sScrMsg *msg, PyObject *msgDict)
	{
		if (static_cast<const sYorNMsg*>(msg)->YorN)
			PyDict_SetItemString(msgDict, "YorN", Py_True);
		else
			PyDict_SetItemString(msgDict, "YorN", Py_False);
	}

	sScrMsg * ScriptMessages::NewKeypadMsg(PyObject *msgDict)
	{
		sKeypadMsg *ret = new sKeypadMsg();
		PyObject *obCode = PyNumber_Int(PyDict_GetItemString(msgDict, "code"));
		if (obCode)
		{
			ret->code = PyInt_AsLong(obCode);
			Py_DECREF(obCode);
		}
		return static_cast<sScrMsg*>(ret);
	}

	void ScriptMessages::NewKeypadObj(const sScrMsg *msg, PyObject *msgDict)
	{
		const sKeypadMsg *m = static_cast<const sKeypadMsg*>(msg);
		PyObject *obCode = PyInt_FromLong(m->code);
		PyDict_SetItemString(msgDict, "code", obCode);
		Py_DECREF(obCode);
	}


} // namespace PyNumbra
